
import PyQt5.QtGui as QtGui

# needed right after QT imports for high-DPI screens
#QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
#QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

import pyqtgraph as pg


class FancyAxis(pg.AxisItem):
  """PyQtGraph AxisItem that allows tick colors different from the grid color,
  a grid background color, and no axis lines. Default aspect is inspired by Seaborn.
  Note that only one axis should draw the background (the other's backgroundColor should be None)"""
  def __init__(self, *args, first=True, backgroundColor="#EAEAF2", tickColor="white", tickWidth=2, **kwargs):
    super(FancyAxis, self).__init__(*args, **kwargs)
    self.backgroundColor = backgroundColor
    self.tickColor = tickColor
    self.tickWidth = tickWidth
    self.z_is_set = False
  

  def drawPicture(self, p, axisSpec, tickSpecs, textSpecs):
    # if this axis is responsible for drawing the background, ensure
    # that the Z value is such that this does not occlude other axes
    if self.backgroundColor and not self.z_is_set:
      self.setZValue(self.zValue() - 1)
      self.z_is_set = True
    
    p.setRenderHint(p.Antialiasing, False)
    p.setRenderHint(p.TextAntialiasing, True)

    # draw background rect
    if self.backgroundColor:
      #bounds = self.mapRectFromParent(self.geometry())
      linkedView = self.linkedView()
      if linkedView is not None and self.grid is not False:
        bounds = linkedView.mapRectToItem(self, linkedView.boundingRect())
        p.fillRect(bounds, QtGui.QColor(self.backgroundColor))

    # draw ticks/grid
    for pen, p1, p2 in tickSpecs:
      pen.setColor(QtGui.QColor(self.tickColor))
      pen.setWidth(self.tickWidth)
      p.setPen(pen)
      p.drawLine(p1, p2)

    # draw all text
    if self.style['tickFont'] is not None:
      p.setFont(self.style['tickFont'])
    p.setPen(self.pen())
    for rect, flags, text in textSpecs:
      p.drawText(rect, flags, text)


def create_plot_widget():
  # create our special axis. only one of them draws the background.
  axis = {'left': FancyAxis('left', backgroundColor=None), 'bottom': FancyAxis('bottom')}
  font = QtGui.QFont()
  font.setPixelSize(14)
  for ax in axis.values():
    ax.tickFont = font
    ax.setStyle(tickTextOffset=10)  # , tickTextWidth=30, tickTextHeight=18  # last ones seem to have no effect

  # create plot widget and activate grid
  plot_widget = pg.PlotWidget(axisItems=axis)
  plot_widget.showGrid(x=True, y=True, alpha=1)

  plot_widget.is_experiment_plot = True  # mark it, mostly used for ordering panels

  return plot_widget





#!/usr/bin/env python

#############################################################################
#
# This file was adapted from Taurus TEP17, but all taurus dependencies were 
# removed so that it works with just pyqtgraph
#
# Just run it and play with the zoom to see how the labels and tick positions 
# automatically adapt to the shown range
#
#############################################################################
# http://taurus-scada.org
#
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
#
#############################################################################

"""
This module provides date-time aware axis
"""

import numpy
#from pyqtgraph import AxisItem
from datetime import datetime, timedelta
from time import mktime

from .plotwidget import FancyAxis

def timestamp(t=None):
    """Convert datetime object to UNIX timestamp (float)"""
    if t is None: t = datetime.now()
    return float(mktime(t.timetuple()))

class DateAxisItem(FancyAxis):  # FancyAxis inherits from pyqtgraph.AxisItem
    """
    A tool that provides a date-time aware axis. It is implemented as an
    AxisItem that interpretes positions as unix timestamps (i.e. seconds
    since 1970).
    The labels and the tick positions are dynamically adjusted depending
    on the range.
    It provides a  :meth:`attachToPlotItem` method to add it to a given
    PlotItem
    """
    
    # Max width in pixels reserved for each label in axis
    _pxLabelWidth = 80

    def __init__(self, *args, **kwargs):
        super(DateAxisItem, self).__init__(*args, **kwargs)
        self._oldAxis = None

    def tickValues(self, minVal, maxVal, size):
        """
        Reimplemented from PlotItem to adjust to the range and to force
        the ticks at "round" positions in the context of time units instead of
        rounding in a decimal base
        """

        maxMajSteps = int(size/self._pxLabelWidth)

        dt1 = datetime.fromtimestamp(minVal)
        dt2 = datetime.fromtimestamp(maxVal)

        dx = maxVal - minVal
        majticks = []

        if dx > 63072001:  # 3600s*24*(365+366) = 2 years (count leap year)
            d = timedelta(days=366)
            for y in range(dt1.year + 1, dt2.year):
                dt = datetime(year=y, month=1, day=1)
                majticks.append(mktime(dt.timetuple()))

        elif dx > 5270400:  # 3600s*24*61 = 61 days
            d = timedelta(days=31)
            dt = dt1.replace(day=1, hour=0, minute=0,
                             second=0, microsecond=0) + d
            while dt < dt2:
                # make sure that we are on day 1 (even if always sum 31 days)
                dt = dt.replace(day=1)
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 172800:  # 3600s24*2 = 2 days
            d = timedelta(days=1)
            dt = dt1.replace(hour=0, minute=0, second=0, microsecond=0) + d
            while dt < dt2:
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 7200:  # 3600s*2 = 2hours
            d = timedelta(hours=1)
            dt = dt1.replace(minute=0, second=0, microsecond=0) + d
            while dt < dt2:
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 1200:  # 60s*20 = 20 minutes
            d = timedelta(minutes=10)
            dt = dt1.replace(minute=(dt1.minute // 10) * 10,
                             second=0, microsecond=0) + d
            while dt < dt2:
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 120:  # 60s*2 = 2 minutes
            d = timedelta(minutes=1)
            dt = dt1.replace(second=0, microsecond=0) + d
            while dt < dt2:
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 20:  # 20s
            d = timedelta(seconds=10)
            dt = dt1.replace(second=(dt1.second // 10) * 10, microsecond=0) + d
            while dt < dt2:
                majticks.append(mktime(dt.timetuple()))
                dt += d

        elif dx > 2:  # 2s
            d = timedelta(seconds=1)
            majticks = range(int(minVal), int(maxVal))

        else:  # <2s , use standard implementation from parent
            return FancyAxis.tickValues(self, minVal, maxVal, size)

        L = len(majticks)
        if L > maxMajSteps:
            majticks = majticks[::int(numpy.ceil(float(L) / maxMajSteps))]

        return [(d.total_seconds(), majticks)]

    def tickStrings(self, values, scale, spacing):
        """Reimplemented from PlotItem to adjust to the range"""
        ret = []
        if not values:
            return []

        if spacing >= 31622400:  # 366 days
            fmt = "%Y"

        elif spacing >= 2678400:  # 31 days
            fmt = "%Y %b"

        elif spacing >= 86400:  # = 1 day
            fmt = "%b/%d"

        elif spacing >= 3600:  # 1 h
            fmt = "%b/%d-%Hh"

        elif spacing >= 60:  # 1 m
            fmt = "%H:%M"

        elif spacing >= 1:  # 1s
            fmt = "%H:%M:%S"

        else:
            # less than 2s (show microseconds)
            # fmt = '%S.%f"'
            fmt = '[+%fms]'  # explicitly relative to last second

        for x in values:
            try:
                t = datetime.fromtimestamp(x)
                ret.append(t.strftime(fmt))
            except ValueError:  # Windows can't handle dates before 1970
                ret.append('')

        return ret

    def attachToPlotItem(self, plotItem):
        """Add this axis to the given PlotItem
        :param plotItem: (PlotItem)
        """
        self.setParentItem(plotItem)
        viewBox = plotItem.getViewBox()
        self.linkToView(viewBox)

        old_axis = plotItem.axes[self.orientation]['item']
        if old_axis:
            old_axis.hide()
            plotItem.layout.removeItem(old_axis)

        plotItem.axes[self.orientation]['item'] = self
        pos = plotItem.axes[self.orientation]['pos']
        plotItem.layout.addItem(self, *pos)
        self.setZValue(-1000)

        self.setFlag(self.ItemNegativeZStacksBehindParent)

    def detachFromPlotItem(self):
        """Remove this axis from its attached PlotItem
        (not yet implemented)
        """
        raise NotImplementedError()  # TODO


if __name__ == '__main__':

    import time
    import sys
    import pyqtgraph as pg
    from PyQt5 import QtWidgets

    app = QtWidgets.QApplication([])

    w = pg.PlotWidget()

    # Add the Date-time axis
    axis = DateAxisItem(orientation='bottom')
    axis.attachToPlotItem(w.getPlotItem())

    # plot some random data with timestamps in the last hour
    now = time.time()
    timestamps = numpy.linspace(now - 3600, now, 100)
    w.plot(x=timestamps, y=numpy.random.rand(100), symbol='o')
    
    w.show()

    sys.exit(app.exec_())




import sys, argparse, logging

#import os
#os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "TRUE"
#os.environ["QT_SCALE_FACTOR"] = "2.5"

import PyQt5.QtWidgets as QtWidgets
import PyQt5.QtCore as QtCore

# needed right after QT imports for high-DPI screens
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

from .window import Window, set_style
from .experiments import Experiments
from .plots import Plots
from .visualizations import Visualizations


def main():
  # parse command line options
  parser = argparse.ArgumentParser()
  parser.add_argument("folder", help="Root folder where experiments are found.")
  #parser.add_argument("-smoothen", default=0, type=float)
  parser.add_argument("-mpl-dpi", default=100, type=int, help="DPI setting for MatPlotLib plots, may be used if text is too big/small (useful for high-DPI monitors).")
  parser.add_argument("--dashes", action='store_true', default=False, help="Cycle through dashes (line) style instead of colors, to distinguish plot lines.")
  parser.add_argument("--force-reopen-files", action='store_true', default=False, help="Slower but more reliable refresh method, useful for remote files.")
  parser.add_argument("-refresh-plots", default=3, type=int, help="Refresh interval for plot updates, in seconds.")
  parser.add_argument("-refresh-new", default=11, type=int, help="Refresh interval for finding new experiments, in seconds.")
  parser.add_argument("-refresh-vis", default=10, type=int, help="Refresh interval for visualizations, in seconds.")
  #parser.add_argument("--no-vis-snapshot", action='store_true', default=False, help="Visualizations are draw using a snapshot of the visualization function, saved with each experiment. This ensures visualizations from old experiments are maintained. Passing this option disables this behavior, which may be useful for debugging.")
  parser.add_argument("-max-hidden-history", default=1000, type=int, help="Maximum number of hidden experiments to remember.")
  parser.add_argument("--debug", action='store_true', default=False, help="Does not suppress exceptions during operation, useful for debugging.")
  parser.add_argument("-loader-log", default='warning', choices=['debug', 'info', 'warning', 'error'], help="Logging verbosity level for experiments loader, from most verbose to least verbose.")
  parser.add_argument("-vis-log", default='warning', choices=['debug', 'info', 'warning', 'error'], help="Logging verbosity level for visualizations, from most verbose to least verbose.")
  parser.add_argument("-plots-log", default='warning', choices=['debug', 'info', 'warning', 'error'], help="Logging verbosity level for plots, from most verbose to least verbose.")
  parser.add_argument("--clear-settings", action='store_true', default=False, help="Clears all stored settings about the GUI state.")
  args = parser.parse_args()


  logging.basicConfig(format="[%(levelname)s] %(message)s")

  # since SSH is a common use case, warn in case fs.sshfs is not installed
  if args.folder.lower().startswith('ssh://'):
    try:
      import fs.sshfs
    except ModuleNotFoundError:
      raise ModuleNotFoundError("To load remote experiments over SSH, the fs.sshfs module must be installed.")


  # create exception and warning hooks, to avoid GUI exiting on uncaught exceptions and suppress warnings
  if not args.debug:
    def trap_exceptions(err_type, err_value, traceback):
      logging.exception('Uncaught exception ' + str(err_type), exc_info=err_value)
    def ignore_qt_warnings(msg_type, msg_log_context, msg_string):
      pass
    sys.excepthook = trap_exceptions
    QtCore.qInstallMessageHandler(ignore_qt_warnings)
  

  # create Qt application
  app = QtWidgets.QApplication(sys.argv)
  set_style(app)

  # create window, experiments and plots holders
  window = Window(window_title='OverBoard - ' + args.folder,
    max_hidden_history=args.max_hidden_history, clear_settings=args.clear_settings)
  
  experiments = Experiments(args.folder, window, args.force_reopen_files,
    args.refresh_plots, args.refresh_new, log_level=args.loader_log)

  plots = Plots(window, args.dashes, log_level=args.plots_log)

  visualizations = Visualizations(window, args.mpl_dpi, args.refresh_vis,
    log_level=args.vis_log)

  window.experiments = experiments

  window.show()
  
  app.exec_()
  #sys.exit(app.exec_())


if __name__ == "__main__":
  main()




import sys

from PyQt5 import QtCore, QtWidgets

# needed right after QT imports for high-DPI screens
#QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
#QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

class FlowLayout(QtWidgets.QLayout):
  """Flowing layout of boxes (inserted left-to-right, top-down to fill space)."""
  def __init__(self, parent=None, margin=-1, hspacing=-1, vspacing=-1):
    super(FlowLayout, self).__init__(parent)
    self._hspacing = hspacing
    self._vspacing = vspacing
    self._items = []
    self.setContentsMargins(margin, margin, margin, margin)

  def __del__(self):
    del self._items[:]

  def clear(self):
    self._items.clear()

  def addItem(self, item):
    self._items.append(item)

  def horizontalSpacing(self):
    if self._hspacing >= 0:
      return self._hspacing
    else:
      return self.smartSpacing(
        QtWidgets.QStyle.PM_LayoutHorizontalSpacing)

  def verticalSpacing(self):
    if self._vspacing >= 0:
      return self._vspacing
    else:
      return self.smartSpacing(
        QtWidgets.QStyle.PM_LayoutVerticalSpacing)

  def count(self):
    return len(self._items)

  def itemAt(self, index):
    if 0 <= index < len(self._items):
      return self._items[index]

  def takeAt(self, index):
    if 0 <= index < len(self._items):
      return self._items.pop(index)

  def expandingDirections(self):
    return QtCore.Qt.Orientations(0)

  def hasHeightForWidth(self):
    return True

  def heightForWidth(self, width):
    return self.doLayout(QtCore.QRect(0, 0, width, 0), True)

  def setGeometry(self, rect):
    super(FlowLayout, self).setGeometry(rect)
    self.doLayout(rect, False)

  def sizeHint(self):
    return self.minimumSize()

  def minimumSize(self):
    size = QtCore.QSize()
    for item in self._items:
      size = size.expandedTo(item.minimumSize())
    left, top, right, bottom = self.getContentsMargins()
    size += QtCore.QSize(left + right, top + bottom)
    return size

  def doLayout(self, rect, testonly):
    left, top, right, bottom = self.getContentsMargins()
    effective = rect.adjusted(+left, +top, -right, -bottom)
    x = effective.x()
    y = effective.y()
    lineheight = 0
    for item in self._items:
      widget = item.widget()
      hspace = self.horizontalSpacing()
      if hspace == -1:
        hspace = widget.style().layoutSpacing(
          QtWidgets.QSizePolicy.PushButton,
          QtWidgets.QSizePolicy.PushButton, QtCore.Qt.Horizontal)
      vspace = self.verticalSpacing()
      if vspace == -1:
        vspace = widget.style().layoutSpacing(
          QtWidgets.QSizePolicy.PushButton,
          QtWidgets.QSizePolicy.PushButton, QtCore.Qt.Vertical)
      nextX = x + item.sizeHint().width() + hspace
      if nextX - hspace > effective.right() and lineheight > 0:
        x = effective.x()
        y = y + lineheight + vspace
        nextX = x + item.sizeHint().width() + hspace
        lineheight = 0
      if not testonly:
        item.setGeometry(
          QtCore.QRect(QtCore.QPoint(x, y), item.sizeHint()))
      x = nextX
      lineheight = max(lineheight, item.sizeHint().height())
    return y + lineheight - rect.y() + bottom

  def smartSpacing(self, pm):
    parent = self.parent()
    if parent is None:
      return -1
    elif parent.isWidgetType():
      return parent.style().pixelMetric(pm, None, parent)
    else:
      return parent.spacing()




from PyQt5 import QtCore, QtWidgets

# needed right after QT imports for high-DPI screens
#QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
#QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

class Slider(QtWidgets.QSlider):
  """Slider that updates its value on mouse drag (not just click)"""
  def mousePressEvent(self, event):
    super(Slider, self).mousePressEvent(event)
    if event.button() == QtCore.Qt.LeftButton:
      val = self.pixelPosToRangeValue(event.pos())
      self.setValue(val)

  def mouseMoveEvent(self, event):
    super(Slider, self).mouseMoveEvent(event)
    val = self.pixelPosToRangeValue(event.pos())
    self.setValue(val)

  def pixelPosToRangeValue(self, pos):
    opt = QtWidgets.QStyleOptionSlider()
    self.initStyleOption(opt)
    gr = self.style().subControlRect(QtWidgets.QStyle.CC_Slider, opt, QtWidgets.QStyle.SC_SliderGroove, self)
    sr = self.style().subControlRect(QtWidgets.QStyle.CC_Slider, opt, QtWidgets.QStyle.SC_SliderHandle, self)

    if self.orientation() == QtCore.Qt.Horizontal:
      sliderLength = sr.width()
      sliderMin = gr.x()
      sliderMax = gr.right() - sliderLength + 1
    else:
      sliderLength = sr.height()
      sliderMin = gr.y()
      sliderMax = gr.bottom() - sliderLength + 1;
    pr = pos - sr.center() + sr.topLeft()
    p = pr.x() if self.orientation() == QtCore.Qt.Horizontal else pr.y()
    return QtWidgets.QStyle.sliderValueFromPosition(self.minimum(), self.maximum(), p - sliderMin, sliderMax - sliderMin, opt.upsideDown)




from PyQt5.QtCore import QThread, QObject, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QMessageBox

import json, re, logging
from time import time
from datetime import datetime

from fs import open_fs, path as fs_path  # pyfilesystem
from fs.errors import FSError

logger = logging.getLogger('overboard.exp')


class Experiments():
  """Stores all Experiment objects, in the name-experiment mapping exps.
  Also manages a thread to find new experiments asynchronously."""
  def __init__(self, base_folder, window, force_reopen_files, poll_time, crawler_poll_time, log_level):
    # set logging messages threshold level
    logger.setLevel(getattr(logging, log_level.upper(), None))

    self.exps = {}
    self.base_folder = base_folder
    self.window = window
    self.force_reopen_files = force_reopen_files
    self.poll_time = poll_time
    
    # create crawler object
    try:
      self.crawler = ExperimentsCrawler(base_folder=base_folder, crawler_poll_time=crawler_poll_time)
    except BaseException as e:
      self.on_error_raised(str(e))
      return  # failed to open file system, no crawling

    # create thread and connect ExperimentsCrawler's signal to Experiments
    # method slot, to return new experiments
    self.thread = QThread()
    self.crawler.experiments_ready.connect(self.on_experiments_ready)
    self.crawler.error_raised.connect(self.on_error_raised)  # same for any errors raised

    self.crawler.moveToThread(self.thread)  # move the crawler object to the thread
    self.thread.started.connect(self.crawler.start_crawling)  # connect thread started signal to crawler slot
    self.thread.start()  # start thread

  def on_experiments_ready(self, filepaths):
    """The crawler found new files, initialize corresponding Experiment objects"""
    for filepath in filepaths:
      exp = Experiment(filepath, self.base_folder, self.force_reopen_files, self.poll_time, self.window)
      self.exps[exp.name] = exp
  
  def on_error_raised(self, msg):
    """The crawler found an error (e.g. no experiments found, or inaccessible path)"""
    QMessageBox.critical(self.window, "OverBoard - Error",
      f'Encountered an error while indexing the experiments path:\n{self.base_folder}\n\nNo more experiments will be indexed. The error was:\n"{msg}"')


class ExperimentsCrawler(QObject):
  """Searches for new experiments asynchronously, on a separate thread"""

  # signal to return new experiments to the Experiments object, on the main thread
  experiments_ready = pyqtSignal(list)

  # to report back errors
  error_raised = pyqtSignal(str)

  def __init__(self, base_folder, crawler_poll_time):
    super().__init__()
    self.base_folder = base_folder
    self.crawler_poll_time = crawler_poll_time
    self.fs = open_fs(base_folder)

  @pyqtSlot()
  def start_crawling(self):
    """Check for new experiments.
    Since the main use case involves remote files mounted with SSHFS/NFS, polling is
    the only viable mechanism to detect changes. This is further argued here:
    https://github.com/samuelcolvin/watchgod#why-no-inotify--kqueue--fsevent--winapi-support"""

    if self.fs is None:
      return
    try:
      known_files = set()
      while True:
        # find experiment files, and see if there are any new ones
        files_exist = False
        new_files = []
        last_time = time()
        for filename in self.fs.walk.files(filter=['stats.csv']):
          files_exist = True
          if filename not in known_files:  # it's new
            new_files.append(filename)

            # if it's taking a while, show an intermediate result
            if time() - last_time > 0.5:  # elapsed time in seconds
              # return the new files as a signal to the main thread
              if len(new_files) > 0:
                self.experiments_ready.emit(new_files)
                known_files.update(set(new_files))
                new_files.clear()
              last_time = time()

        if not files_exist:
          raise IOError("The directory does not contain any experiments (stats.csv files).")

        # return the new files as a signal to the main thread
        if len(new_files) > 0:
          self.experiments_ready.emit(new_files)
          known_files.update(set(new_files))
          new_files.clear()

        # wait some time before looking again
        QThread.sleep(self.crawler_poll_time)

    # if anything bad happens, pop up a message box and stop crawling
    except BaseException as e:
      self.error_raised.emit(str(e))


class Experiment():
  """Stores data for a single experiment, and manages a thread to read new data asynchronously"""
  def __init__(self, filepath, base_folder, force_reopen_files, poll_time, window):
    (rel_path, _, filename) = filepath.rpartition('/')  # extract '/stats.csv' from path; forward slash guaranteed by pyfilesystem
    self.name = rel_path[1:]  # the experiment name is the directory path, relative to the base folder
    self.filename = filename
    self.directory = fs_path.combine(base_folder, rel_path)

    logger.debug(f"Initializing thread to load {self.name}")

    self.meta = {}
    self.metrics = []  # names of metrics
    self.data = []  # data for each metric (one list per metric)
    self.done = False  # true after reading and the experiment is done writing too

    # start hidden if the user hid it the last time (this is a persistent setting)
    if self.directory in window.hidden_exp_paths:
      self.visible = False
      del window.hidden_exp_paths[self.directory]
    else:
      self.visible = True

    self.is_selected = False
    self.is_filtered = False
    self.style_idx = None  # style index, for when it's assigned (see Plots.assign_exp_style)

    # register this experiment with the main window
    self.window = window
    self.table_row = None  # used internally by the window
    window.on_exp_init(self)

    # create reader object and thread
    self.reader = ExperimentReader(filename=filename, directory=self.directory, force_reopen_files=force_reopen_files, poll_time=poll_time, name=self.name)
    self.thread = QThread()

    # connect ExperimentReader's signals to Experiment method slots, to return data
    self.reader.meta_ready.connect(self.on_meta_ready)
    self.reader.header_ready.connect(self.on_header_ready)
    self.reader.data_ready.connect(self.on_data_ready)
    self.reader.done.connect(self.on_done)

    self.reader.moveToThread(self.thread)  # move the reader object to the thread

    self.reader.done.connect(self.thread.quit)  # connect reader done signal to terminate thread slot
    self.thread.started.connect(self.reader.start_reading)  # connect thread started signal to reader slot

    self.thread.start()  # start thread


  # receive signals from thread, and store associated data
  def on_meta_ready(self, meta):  
    self.meta = meta
    self.window.on_exp_meta_ready(self)

  def on_header_ready(self, header):  
    self.metrics = header
    self.data = [[] for _ in header]  # initialize each column of data
    self.window.on_exp_header_ready(self)

  def on_data_ready(self, data):  
    assert(len(self.metrics) > 0)  # sanity check, on_header_ready should have been called before

    # append new values to each existing column, and update plots
    for (column, new_values) in zip(self.data, data):
      column.extend(new_values)

    self.window.plots.add(self)

  def on_done(self):
    # mark experiment as done. this signal is also sent to the QThread's quit slot, so it ends.
    self.done = True

  def is_visible(self):
    # helper for whether it's visible and not filtered out
    return (self.visible and not self.is_filtered)



class ExperimentReader(QObject):
  """Reads data from an experiment asynchronously, on a separate thread"""

  # signals to return data to the Experiment object, on the main thread
  meta_ready = pyqtSignal(dict)
  header_ready = pyqtSignal(list)
  data_ready = pyqtSignal(list)
  done = pyqtSignal()

  def __init__(self, filename, directory, force_reopen_files, poll_time, name):
    super().__init__()
    self.name = name
    self.filename = filename
    self.directory = directory
    self.force_reopen_files = force_reopen_files
    self.poll_time = poll_time
    self.num_columns = None
    self.num_rows = 0

    self.fs = open_fs(directory)

  @pyqtSlot()
  def start_reading(self):
    """Do all the reading asynchronously from the main thread"""

    # read JSON file with metadata (including timestamp), if it exists
    meta = {}
    try:
      meta = json.loads(self.fs.readtext('meta.json'))
    except (IOError, FSError):  # no meta data, not critical
      self.meta_ready.emit({})

    # try interpreting string values in meta-data as an ISO date
    for (key, value) in meta.items():
      if isinstance(value, str):
        try:  # try interpreting as an ISO date
          meta[key] = datetime.fromisoformat(value)
        except (ValueError, AttributeError):
          pass  # otherwise, keep as a string

    # send the meta-data to the main thread
    self.meta_ready.emit(meta)

    # read data
    logger.debug(f"Finished reading experiment meta-data; now reading CSV for {self.name}")
    try:
      if self.force_reopen_files:
        self.read_data_slow()
      else:
        self.read_data_fast()
    except (IOError, FSError, ValueError):
      logger.exception(f"Error reading {self.directory}/{self.filename}")
  
  def read_data_fast(self):
    """Polls the file for new data by attempting another read, keeping the file handle open (fast)"""
    done = False
    with self.fs.open(self.filename) as file:
      (done, line_start) = self.read_lines(file)
      while not done:
        # try to read new data after a while. need to restore read position when there's
        # no new line to read, to cope with incomplete lines (while they're being written)
        QThread.sleep(self.poll_time)
        file.seek(line_start)
        (done, line_start) = self.read_lines(file)

  def read_data_slow(self):
    """Polls the file for new data by tracking size changes, and reopening it to read each time (slower)"""
    old_size = 0
    line_start = None
    while True:
      info = self.fs.getinfo(self.filename, namespaces=['details'])

      if info.size != old_size:
        old_size = info.size
        with self.fs.open(self.filename, 'r') as file:
          # get new data from the file, picking up where we left off. don't use file size, depends on OS
          if line_start is not None:
            file.seek(line_start)

          (done, line_start) = self.read_lines(file)
          if done: return

      QThread.sleep(self.poll_time)

  def read_lines(self, file):
    """Reads as many lines as possible from an open CSV file, emitting appropriate signals with the data and status.
    Returns whether the experiment is finished (boolean), and the position to re-start reading from."""
    (rows, done) = ([], False)
    logger.debug(f"Starting to read lines for {self.name}")

    while True:
      # try to read a new line. need to restore read position when there's no new
      # line to read, to cope with incomplete lines (while they're being written)
      line_start = file.tell()
      line = file.readline()
      
      # an empty line, terminated by a line break (\n, \r or \r\n), marks the end
      # of the experiment; no further reading necessary
      if 1 <= len(line) <= 2 and (char in ('\n', '\r') for char in line):
        done = True
        logger.debug(f"End of experiment {self.name}")
        break

      # reached the end of file, but there's no line break; writing is not complete
      if len(line) == 0 or line[-1] not in ('\n', '\r'):
        logger.debug(f"Incomplete line for {self.name}")
        break

      if self.num_columns is None:
        # read CSV file header (with stat names). they're separated by commas, which can be escaped: \,
        headers = re.split(r'(?<!\\),', line.strip())
        
        if len(headers) == 0: raise IOError('CSV file has too few headers.')

        headers.insert(0, 'iteration')  # insert a default column, the row number
        self.num_columns = len(headers)

        # send a signal with the headers
        self.header_ready.emit(headers)

        logger.debug(f"Read headers for {self.name}: {headers}")
      else:
        # interpret line of stat values
        #row = [float(v) for v in line.split(',')]
        row = [self.num_rows]  # start with iteration count (row number)
        for value in line.split(','):
          try:  # first try converting to float
            value = float(value)
          except ValueError:
            try:  # try interpreting as an ISO date
              value = datetime.fromisoformat(value)
            except (ValueError, AttributeError):
              pass  # otherwise, keep as a string
          row.append(value)

        if len(row) != self.num_columns:
          print(self.num_columns, row)
          raise IOError('A CSV line has an incorrect number of values (compared to the header).')
        
        rows.append(row)
        self.num_rows += 1
        logger.debug(f"Read row {self.num_rows} for {self.name}")

    # send signals that data is ready or the experiment is done, if needed.
    # also return position to seek to, to re-read any incomplete line.
    if len(rows) > 0:  # note we transpose the data (columns instead of rows)
      columns = list(zip(*rows))
      self.data_ready.emit(columns)
    if done:
      self.done.emit()
    return (done, line_start)





import PyQt5.QtCore as QtCore

# needed right after QT imports for high-DPI screens
#QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
#QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

from functools import partial
import heapq, logging
from datetime import datetime
from numbers import Number
from itertools import zip_longest
from random import random
import numpy as np

import pyqtgraph as pg

from .plotwidget import create_plot_widget
from .pg_time_axis import timestamp, DateAxisItem

# define lists of styles to cycle
palette = ["#4C72B0", "#DD8452", "#55A868", "#C44E52", "#8172B3", "#937860", "#DA8BC3", "#8C8C8C", "#CCB974", "#64B5CD"]
dashes = [QtCore.Qt.SolidLine, QtCore.Qt.DashLine, QtCore.Qt.DotLine, QtCore.Qt.DashDotLine, QtCore.Qt.DashDotDotLine]
#dashes_by_name = dict(zip(['-', '--', ':', '-.', '-..'], dashes))
widths = [2, 1, 3]  # line widths

logger = logging.getLogger('overboard.plt')


class Plots():
  def __init__(self, window, dashes, log_level):
    # set logging messages threshold level
    logger.setLevel(getattr(logging, log_level.upper(), None))

    self.window = window
    window.plots = self  # back-reference
    self.dashes = dashes  # option to cycle through dashes first instead of colors

    self.panels = {}  # widgets containing plots, indexed by name (usually the plot title at the top)
    self.hovered_plot_info = None

    # reuse styles from hidden experiments if possible (early styles are
    # more distinguishable). note we only store style indexes, to allow
    # easy changing later (e.g. style by dashes only or colors).
    self.unused_styles = []  # keep unused styles in a heap so early styles have priority
    self.next_style_index = 0  # next unused style that is not in the heap

    # create timer to restore auto-range of plot axis progressively
    # (auto-range is disabled when plotting for performance)
    self.autorange_panels = {}
    self.autorange_timer = QtCore.QTimer()
    self.autorange_timer.timeout.connect(self.restore_autorange)
    self.autorange_timer.start(500)
    
    # set general PyQtGraph options
    pg.setConfigOptions(antialias=True, background='w', foreground='k')  # black on white

  def assign_exp_style(self, exp):
    """Assign a new style to an experiment"""
    # reuse a previous style if possible, in order
    if len(self.unused_styles) > 0:
      exp.style_idx = heapq.heappop(self.unused_styles)
    else:
      # otherwise, get a new one
      exp.style_idx = self.next_style_index
      self.next_style_index += 1

  def drop_exp_style(self, exp):
    """Remove the style of an experiment, and consider that style available for others"""
    if exp.style_idx is not None:
      heapq.heappush(self.unused_styles, exp.style_idx)
      exp.style_idx = None
  
  def drop_all_exp_styles(self):
    """Quickly reset the styles of all experiments"""
    self.unused_styles.clear()
    self.next_style_index = 0
    for exp in self.window.experiments.exps.values():
      exp.style_idx = None

  def get_exp_style(self, exp, assign=True):
    """Return a dict with the unique style of an experiment, assigning one if necessary"""
    if exp.style_idx is None:
      if not assign: return None  # asked to not assign a style
      self.assign_exp_style(exp)
    idx = exp.style_idx

    # start by varying color, then dashes, then line width
    if not self.dashes:
      color = palette[idx % len(palette)]
      dash = dashes[(idx // len(palette)) % len(dashes)]
      width = widths[(idx // (len(palette) * len(dashes))) % len(widths)]
    else:
      # alternative order: dashes, width, color
      dash = dashes[idx % len(dashes)]
      width = widths[(idx // len(dashes)) % len(widths)]
      color = palette[(idx // (len(dashes) * len(widths))) % len(palette)]

    return {'color': color, 'style': dash, 'width': width}

  def define_plots(self, exp):
    """Defines plot information for a given experiment. Returns a list of dicts,
    each dict defining the properties of a single line (sources for x, y, color,
    panel, etc). This accesses the window widgets to check the current options."""

    x_option = self.window.x_dropdown.currentText()
    y_option = self.window.y_dropdown.currentText()
    panel_option = self.window.panel_dropdown.currentText()
    merge_option = self.window.merge_dropdown.currentText()
    metrics_subset = set(self.window.metrics_subset_dropdown.get_checked_list())

    # some combinations are invalid, change to sensible defaults in those cases
    if panel_option == "One per metric":
      # Panel metric is unused, so all panels would look the same
      if x_option != "Panel metric" and y_option != "Panel metric":
        y_option = "Panel metric"
        self.window.y_dropdown.setCurrentText(y_option)
    
    else:
      # x/y_option "Panel metric" is only compatible with panel_option "One per metric"
      if x_option == "Panel metric":
        x_option = "iteration"
        self.window.x_dropdown.setCurrentText(x_option)
      if y_option == "Panel metric":
        y_option = "time"
        self.window.y_dropdown.setCurrentText(y_option)

    # don't let X and Y axis be set to the same value
    if x_option == y_option:
      if x_option != "iteration":
        x_option = "iteration"
      else:
        x_option = "time (relative)"
      self.window.x_dropdown.setCurrentText(x_option)

    # create list of panels by: metric, experiment, hyper-parameter type,
    # value of a single hyper-parameter, or create a single panel
    if panel_option == "One per metric":
      panels = [m for m in exp.metrics if m != "iteration" and m != "time"]
    elif panel_option == "One per run":
      panels = [exp.name]
    elif panel_option == "Single panel":
      panels = [y_option]
    else:  # single hyper-parameter selected, create different panels based on value
      panels = [panel_option + ' = ' + str(exp.meta.get(panel_option, None))]  # None if missing

    # possibly merge lines by some hyper-parameter; otherwise, each experiment is unique
    if merge_option == "Nothing" or merge_option not in exp.meta:
      (merge_info, line_id) = (None, exp.name)
    else:
      (merge_info, line_id) = (exp.name, str(exp.meta[merge_option]))

    info = []  # the list of lines to plot
    for panel in panels:  # possibly spread plots across panels
      logger.debug(f"Panel {panel}")

      lines = [(x_option, y_option)]  # single line per panel, with these X and Y sources
      if x_option == "All metrics":  # several lines in a panel - one per metric
        lines = [(name, y_option) for name in exp.metrics]
      elif y_option == "All metrics":
        lines = [(x_option, name) for name in exp.metrics]

      for (x, y) in lines:  # possibly create multiple lines for this panel
        logger.debug(f"Line x={x}, y={y}")
        if x_option == "Panel metric": x = panel  # different metrics per panel
        if y_option == "Panel metric": y = panel

        # special option: time aligned to the same origin
        if x == 'time (relative)': (x, x_relative) = ('time', True)
        else: x_relative = False
        if y == 'time (relative)': (y, y_relative) = ('time', True)
        else: y_relative = False

        # skip if not part of the subset
        if ((x not in exp.meta and x not in metrics_subset) or
            (y not in exp.meta and y not in metrics_subset)):
          logger.debug("Skipping since x or y not in metrics subset")
          continue

        # a plot with the same values on X and Y is redundant, so skip it
        if x == y:
          logger.debug("Skipping since x==y")
          continue

        # skip if this experiment does not have the required data
        if x not in exp.meta and x not in exp.metrics:
          logger.debug("Skipping since x not in meta or metrics")
          continue
        
        if y not in exp.meta and y not in exp.metrics:
          logger.debug("Skipping since y not in meta or metrics")
          continue

        # final touches and compose dict
        info.append(dict(panel=panel, x=x, y=y, line_id=(x, y, line_id),
          exp_name=exp.name, merge_info=merge_info,
          x_relative=x_relative, y_relative=y_relative))
    return info

  def add(self, exp):
    """Creates or updates plots associated with given experiment, creating panels if needed.
    If the experiment is marked as invisible/filtered, nothing will be drawn."""

    if not exp.is_visible() or len(exp.metrics) == 0:
      return False  # plots are invisible or no data loaded yet

    logger.debug(f"Adding plots from experiment {exp.name}")

    plots = self.define_plots(exp)
    for plot in plots:
      # create new panel if it doesn't exist
      if plot['panel'] not in self.panels:
        self.add_plot_panel(plot)

      panel = self.panels[plot['panel']]  # reuse existing panel
      self.pause_autorange(panel)
      plot_item = panel.plot_widget.getPlotItem()
      
      # get data points, pre-processed to ensure they are numeric. this may edit the axes.
      (xs, ys, x_is_categ, y_is_categ) = self.get_numeric_data_points(exp, plot, plot_item)

      # check if plot line already exists
      if plot['line_id'] not in panel.plots_dict:
        # create new line
        line = plot_item.plot([], [])
        line.curve.setClickable(True, 8)  # size of hover region
        panel.plots_dict[plot['line_id']] = line
      else:
        # update existing one
        line = panel.plots_dict[plot['line_id']]
      line.plot_info = plot  # store the plot information for later, e.g. on mouse-over
      line.mouse_over = False

      has_new_style = (exp.style_idx is None)  # remember if a new style is assigned

      if plot['merge_info'] is not None:
        # handle merged plots, by updating the statistics to display first
        (xs, ys, shade_y1, shade_y2) = self.update_merged_stats(line, plot['merge_info'], xs, ys)

        # share the same style among a group of merged experiments
        if exp.style_idx is None:
          if not hasattr(line, 'style_idx'):
            self.assign_exp_style(exp)  # new style
            line.style_idx = exp.style_idx
          else:  # use the same style as the previous merged experiments
            exp.style_idx = line.style_idx

      # get the experiment's style (color, dashes, etc)
      style = self.get_exp_style(exp)

      if has_new_style:  # update the icon if the style was missing before
        self.window.redraw_icon(exp)
      
      if exp.is_selected:  # selected lines are thicker
        style['width'] = style.get('width', 2) + 2
      
      # create pen with the experiment's style, and args to assign to PlotDataItem line
      pen = pg.mkPen(style)
      data = dict(x=xs, y=ys, pen=pen)

      # for single points, plot a marker/symbol, since the line won't show up
      if len(xs) == 1:
        data['symbol'] = 'o'
        data['symbolBrush'] = pen.color()
        data['symbolSize'] = pen.width() * 2 + 4
        
        # for categorical axis, jitter single points. unfortunately, points will
        # jump around when selecting/deselecting plots, so we need to keep the
        # amount of jitter in a state variable per line.
        if x_is_categ:
          if not hasattr(line, 'jitter_x'): line.jitter_x = random() * 0.2 - 0.1
          xs[0] += line.jitter_x
        if y_is_categ:
          if not hasattr(line, 'jitter_y'): line.jitter_y = random() * 0.2 - 0.1
          ys[0] += line.jitter_y
      else:
        data['symbol'] = None

      # assign the point coordinates and visual properties to the PlotDataItem
      line.setData(**data)

      # finish merged plots, by plotting the confidence intervals
      if plot['merge_info'] is not None:
        if len(xs) > 1:
          # draw a shaded area. first, set the pen used to draw the outline of the shaded area
          outline_pen = pg.mkPen(pen)
          outline_pen.setWidthF(pen.widthF() / 3)
          
          if plot['line_id'] not in panel.aux_plots_dict:
            # create for first time. we need 2 curves, setting the upper and lower
            # limits, and then a FillBetweenItem to shade the space between them.
            limit1 = plot_item.plot([], [])
            limit2 = plot_item.plot([], [])
            shade = pg.FillBetweenItem(limit1, limit2, (200, 0, 0, 128))
            plot_item.addItem(shade)
            panel.aux_plots_dict[plot['line_id']] = (limit1, limit2, shade)
          else:
            (limit1, limit2, shade) = panel.aux_plots_dict[plot['line_id']]
          limit1.setData(x=xs, y=shade_y1, pen=outline_pen)
          limit2.setData(x=xs, y=shade_y2, pen=outline_pen)

          c = pen.color()  # shade using same color but semi-transparent
          shade.setBrush((c.red(), c.green(), c.blue(), 64))
        else:
          # a single point, plot as an error bar
          data = dict(x=xs, y=ys, bottom=ys-shade_y1, top=shade_y2-ys, pen=pen)
          if plot['line_id'] not in panel.aux_plots_dict:
            # create for first time
            bar = panel.aux_plots_dict[plot['line_id']] = pg.ErrorBarItem(**data)
            plot_item.addItem(bar)
          else:
            panel.aux_plots_dict[plot['line_id']].setData(**data)

    return len(plots) > 0  # True if some plots were actually drawn

  
  def add_plot_panel(self, plot):
    """Adds a single plot panel, from its description as output
    by define_plots. Used by Plots.add."""

    # create new panel to contain plot
    title = plot['panel']
    plot_widget = create_plot_widget()
    panel = self.window.add_panel(plot_widget, title)

    logger.debug(f"Adding plot panel {title}")

    plot_item = panel.plot_widget.getPlotItem()
    plot_item.setLabel('bottom', plot['x'])  # set X axis label

    # mouse cursor (vertical line)
    panel.cursor_vline = pg.InfiniteLine(angle=90, pen="#B0B0B0")
    panel.cursor_vline.setVisible(False)
    panel.cursor_vline.setZValue(10)
    plot_item.addItem(panel.cursor_vline, ignoreBounds=True)  # ensure it doesn't mess autorange

    # mouse cursor (dot)
    panel.cursor_dot = pg.PlotDataItem([], [], pen=None, symbolPen=None, symbolBrush="#C00000", symbol='o', symbolSize=7)
    panel.cursor_dot.setVisible(False)
    panel.cursor_dot.setZValue(10)
    plot_item.addItem(panel.cursor_dot, ignoreBounds=True)

    # mouse cursor text
    panel.cursor_label = pg.LabelItem(justify='left')
    panel.cursor_label.setParentItem(plot_item.getViewBox())
    panel.cursor_label.anchor(itemPos=(0, 0), parentPos=(0, 0))
    panel.cursor_label.setZValue(10)
    
    # set up mouse events
    plot_widget.mouseMoveEvent = partial(self.on_mouse_move, panel=panel)
    plot_widget.leaveEvent = partial(self.on_mouse_leave, panel=panel)
    plot_widget.mousePressEvent = partial(self.on_mouse_click, panel=panel)

    panel.plots_dict = {}
    panel.aux_plots_dict = {}
    self.panels[plot['panel']] = panel

  def get_numeric_data_points(self, exp, plot, plot_item):
    """Retrieves the data points for a single plot, from its define_plots
    description. Reduction to scalar (single point) is applied if needed, and
    time/categorical data is converted to numeric coordinates. Used by Plots.add."""
    (x_is_categ, y_is_categ) = (False, False)  # whether an axis is categorical

    if plot['x'] in exp.meta:
      xs = [exp.meta[plot['x']]]  # a single point, with the chosen hyper-parameter
    else:
      if plot['x'] not in exp.metrics:  # final sanity check
        logging.warning("The chosen metric was not found in this experiment.")
        xs = []
      else:
        xs = exp.data[exp.metrics.index(plot['x'])]  # several points, with the chosen metric

    if plot['y'] in exp.meta:
      ys = [exp.meta[plot['y']]]
    else:
      if plot['y'] not in exp.metrics:  # final sanity check
        logging.warning("The chosen metric was not found in this experiment.")
        ys = []
      else:
        ys = exp.data[exp.metrics.index(plot['y'])]

    # if one axis is a scalar (hyper-parameter) and another is not (metric), only show
    # a single data point. use "scalar display" option to decide which metric to keep.
    if len(xs) == 1 or len(ys) == 1:
      scalar_option = self.window.scalar_dropdown.currentText()
      if scalar_option == 'Last value':
        if len(xs) > 1: xs = [xs[-1]]
        if len(ys) > 1: ys = [ys[-1]]
      elif scalar_option == 'Maximum':
        if len(xs) > 1: xs = [max(xs)]
        if len(ys) > 1: ys = [max(ys)]
      elif scalar_option == 'Minimum':
        if len(xs) > 1: xs = [min(xs)]
        if len(ys) > 1: ys = [min(ys)]

    assert len(xs) == len(ys)


    # check data points' types to know what axes to create (numeric, time or categorical).
    # handle datetimes. create time axes if needed, and convert datetimes to numeric values
    if len(xs) > 0 and all(isinstance(x, datetime) for x in xs):
      if not isinstance(plot_item.axes['bottom']['item'], DateAxisItem):
        axis = DateAxisItem(orientation='bottom')
        axis.attachToPlotItem(plot_item)
        axis.setGrid(255)

      # plot values relative to the same origin (i.e. remove minimum)
      if plot['x_relative']:
        earliest = min(xs)
        origin = datetime(year=2000, month=1, day=1)
        xs = [x - earliest + origin for x in xs]

      xs = [timestamp(x) for x in xs]

    # handle categorical values
    elif len(xs) > 0 and (self.window.x_categorical_checkbox.isChecked() or
     any(not isinstance(x, Number) or isinstance(x, bool) for x in xs)):
      axes = plot_item.axes['bottom']['item']
      if axes._tickLevels is None:  # initialize
        axes.setTicks([[]])
        axes.ticks_dict = {}
        axes.next_tick = 0
      ticks_dict = axes.ticks_dict

      # TODO: improve performance
      xs = [str(x) for x in xs]  # ensure they're all strings
      for x in set(xs):  # iterate unique values
        if x not in axes.ticks_dict:  # add tick if this value is new
          ticks_dict[x] = axes.next_tick
          axes._tickLevels[0].append((axes.next_tick, x))
          axes.next_tick += 1

      # convert to numeric value, by look-up
      xs = [ticks_dict[x] for x in xs]
      x_is_categ = True


    # same as above, for Y axis.
    # handle datetimes. create time axes if needed, and convert datetimes to numeric values
    if len(ys) > 0 and all(isinstance(y, datetime) for y in ys):
      if not isinstance(plot_item.axes['left']['item'], DateAxisItem):
        axis = DateAxisItem(orientation='left', backgroundColor=None)
        axis.attachToPlotItem(plot_item)
        axis.setGrid(1)

      # plot values relative to the same origin (i.e. remove minimum)
      if plot['y_relative']:
        earliest = min(ys)
        origin = datetime(year=2000, month=1, day=1)
        ys = [y - earliest + origin for y in ys]

      ys = [timestamp(y) for y in ys]

    # handle categorical values
    elif len(ys) > 0 and (self.window.y_categorical_checkbox.isChecked() or
     any(not isinstance(y, Number) or isinstance(y, bool) for y in ys)):

      axes = plot_item.axes['left']['item']
      if axes._tickLevels is None:  # initialize
        axes.setTicks([[]])
        axes.ticks_dict = {}
        axes.next_tick = 0
      ticks_dict = axes.ticks_dict

      ys = [str(y) for y in ys]  # ensure they're all strings
      for y in set(ys):  # iterate unique values
        if y not in axes.ticks_dict:  # add tick if this value is new
          ticks_dict[y] = axes.next_tick
          axes._tickLevels[0].append((axes.next_tick, y))
          axes.next_tick += 1
          
      # convert to numeric value, by look-up
      ys = [ticks_dict[y] for y in ys]
      x_is_categ = True
    
    return (xs, ys, x_is_categ, y_is_categ)

  def update_merged_stats(self, line, merge_info, xs=None, ys=None):
    # update the unmerged data, stored in the line object
    if not hasattr(line, 'unmerged_xs'):
      line.unmerged_xs = {}
      line.unmerged_ys = {}
    
    if xs is not None:  # add it
      line.unmerged_xs[merge_info] = xs
    else:  # deleting this entry
      del line.unmerged_xs[merge_info]
      if len(line.unmerged_xs) == 0:  # nothing left
        return [None] * 4

    if ys is not None:
      line.unmerged_ys[merge_info] = ys
    else:
      del line.unmerged_ys[merge_info]
      if len(line.unmerged_xs) == 0:
        return [None] * 4

    # convert all data to numpy arrays, shape = (max number of points, number of repeats)
    # zip_longest is needed to fill in missing values with NaN, for incomplete lines.
    all_xs = np.array(list(zip_longest(*list(line.unmerged_xs.values()), fillvalue=np.nan)))
    all_ys = np.array(list(zip_longest(*list(line.unmerged_ys.values()), fillvalue=np.nan)))

    # compute statistics      
    merged_line = self.window.merge_line_dropdown.currentText()
    merged_shade = self.window.merge_shade_dropdown.currentText()

    if merged_line == 'Median':
      xs = np.nanmedian(all_xs, axis=1, keepdims=False)
      ys = np.nanmedian(all_ys, axis=1, keepdims=False)
    else:  # mean
      xs = np.nanmean(all_xs, axis=1, keepdims=False)
      ys = np.nanmean(all_ys, axis=1, keepdims=False)

    if merged_shade == 'Maximum and minimum':
      shade_y1 = np.nanmin(all_ys, axis=1, keepdims=False)
      shade_y2 = np.nanmax(all_ys, axis=1, keepdims=False)
    else:
      # '2 x standard deviations', extract the integer factor in the first character and use it
      factor = int(merged_shade[0])
      std = factor * np.nanstd(all_ys, axis=1, keepdims=False)
      (shade_y1, shade_y2) = (ys - std, ys + std)

    return (xs, ys, shade_y1, shade_y2)

  def remove(self, exp):
    """Removes all plots associated with an experiment (inverse of Plots.add)"""
    if len(exp.metrics) == 0:  # no data yet
      return

    plots = self.define_plots(exp)
    for plot in plots:
      # find panel
      if plot['panel'] in self.panels:
        panel = self.panels[plot['panel']]
        self.pause_autorange(panel)

        # find plot line
        line_id = plot['line_id']
        if line_id in panel.plots_dict:
          line = panel.plots_dict[line_id]

          if plot['merge_info'] is not None:
            # remove this data from the merged line, and update it
            (xs, ys, shade_y1, shade_y2) = self.update_merged_stats(line, plot['merge_info'])
            if xs is not None:
              (limit1, limit2, shade) = panel.aux_plots_dict[plot['line_id']]
              limit1.setData(x=xs, y=shade_y1)
              limit2.setData(x=xs, y=shade_y2)
            else:
              # removed final line, nothing left
              plot['merge_info'] = None
              
              # remove auxiliary plots (e.g. shaded merged plots)
              if line_id in panel.aux_plots_dict:
                plot_item = panel.plot_widget.getPlotItem()
                for aux_object in panel.aux_plots_dict[line_id]:
                  plot_item.removeItem(aux_object)
                del panel.aux_plots_dict[line_id]
          
          if plot['merge_info'] is None:
            # simple line, remove it
            plot_item = panel.plot_widget.getPlotItem()
            plot_item.removeItem(line)
            del panel.plots_dict[line_id]

        # if the last line was deleted, delete the panel too
        if len(panel.plots_dict) == 0:
          panel.setParent(None)
          panel.deleteLater()
          del self.panels[plot['panel']]

  def remove_all(self):
    """Remove all plots, fast"""
    for panel in self.panels.values():
      panel.setParent(None)
      panel.deleteLater()
    self.window.flow_layout.clear()
    self.panels.clear()


  def on_mouse_move(self, event, panel):
    """Select curves when hovering them, and update mouse cursor text"""
    # access PlotItem's ViewBox to map mouse to data coordinates
    plot_item = panel.plot_widget.getPlotItem()
    point = plot_item.vb.mapSceneToView(event.pos())
    
    hovered = None
    for line in panel.plots_dict.values():
      # only the first one gets selected
      inside = (not hovered and (line.curve.mouseShape().contains(point) or line.scatter.pointsAt(point)))

      if inside:
        hovered = line
        if not line.mouse_over:
          # change line style to thicker
          line.original_pen = line.opts['pen']
          pen = pg.mkPen(line.original_pen)
          pen.setWidthF(line.original_pen.widthF() + 2)
          line.setPen(pen)

          # bring it to the front
          line.setZValue(1)

          line.mouse_over = True
      else:
        if line.mouse_over:
          # restore line style and z-order
          line.setPen(line.original_pen)
          line.setZValue(0)
          line.mouse_over = False

    # show cursor
    panel.cursor_vline.setVisible(True)
    x = point.x()

    if hovered:
      # snap vertical line to nearest point (by x coordinate)
      data = hovered.getData()
      index = np.argmin(np.abs(data[0] - x))
      (x, y) = (data[0][index], data[1][index])

      # snap dot to nearest point too
      panel.cursor_dot.setVisible(True)
      panel.cursor_dot.setData([x], [y])
      vline_x = x

      # show X value as string for categorical axes
      axes = plot_item.axes['bottom']['item']
      if hasattr(axes, 'ticks_dict'):  # X axis is categorical
        index = round(x)  # round due to possible jittering
        # awkward indexing into axes.ticks_dict by value (index/x value) instead of key (text label)
        labels = [label for (label, idx) in axes.ticks_dict.items() if idx == index]
        if len(labels) > 0: x = labels[0]
      else:
        # numeric value. print floats with 3 significant digits and no
        # sci notation (e.g. 1e-4). also consider integers.
        if x % 1 == 0: x = str(int(x))
        else: x = float('%.3g' % x)

      # show Y value as string for categorical axes
      axes = plot_item.axes['left']['item']
      if hasattr(axes, 'ticks_dict'):  # X axis is categorical
        index = round(y)  # round due to possible jittering
        # awkward indexing into axes.ticks_dict by value (index/x value) instead of key (text label)
        labels = [label for (label, idx) in axes.ticks_dict.items() if idx == index]
        if len(labels) > 0: y = labels[0]
      else:
        # numeric value, same as above
        if y % 1 == 0: y = str(int(y))
        else: y = float('%.3g' % y)
      
      # show data coordinates and line information, and store it
      # for on_mouse_click to access later
      info = self.hovered_plot_info = hovered.plot_info
      text = f"{info['exp_name']}<br/>({info['x']}={x}, {info['y']}={y})"

    else:
      panel.cursor_dot.setVisible(False)
      text = ""
      vline_x = x
      self.hovered_plot_info = None

    # set positions and text
    panel.cursor_label.setText(text)  #, size='10pt'
    panel.cursor_vline.setValue(vline_x)

    pg.PlotWidget.mouseMoveEvent(panel.plot_widget, event)

  def on_mouse_leave(self, event, panel):
    """Hide cursor when the mouse leaves"""
    panel.cursor_vline.setVisible(False)
    panel.cursor_dot.setVisible(False)

  def on_mouse_click(self, event, panel):
    """Select experiment associated with the currently hovered line (by name)"""
    if self.hovered_plot_info is not None:
      self.window.select_experiment(self.hovered_plot_info['exp_name'])
      event.accept()
    pg.PlotWidget.mousePressEvent(panel.plot_widget, event)

  def pause_autorange(self, panel):
    """Pause auto-range temporarily when plotting, for performance (restored by a timer)"""
    if panel not in self.autorange_panels:  # if already paused do nothing
      view = panel.plot_widget.getPlotItem().getViewBox()
      state = view.autoRangeEnabled()
      if any(state):  # list with 2 booleans, True if each axis has auto-range enabled
        view.disableAutoRange()
        self.autorange_panels[panel] = state

  def restore_autorange(self):
    """Restore auto-range of plot axis progressively, called by a timer"""
    if len(self.autorange_panels) > 0:
      # pop oldest panel off the stack (this is still an ordered dict to check membership)
      (panel, state) = next(iter(self.autorange_panels.items()))
      del self.autorange_panels[panel]

      view = panel.plot_widget.getPlotItem().getViewBox()
      try:
        view.enableAutoRange(x=state[0], y=state[1])
      except RuntimeError:  # sometimes the object was deleted in the meanwhile
        pass


class Smoother():
  def __init__(self, bandwidth, half_window=None):
    if bandwidth == 0:
      self.kernel = None
    else:
      if half_window is None:
        half_window = int(np.ceil(bandwidth * 2))
      self.kernel = np.exp(-np.arange(-half_window, half_window + 1)**2 / bandwidth**2)
    self.changed = True

  def do(self, x):
    if not isinstance(x, np.ndarray):
      x = np.array(x)
    if self.kernel is None or len(x) == 0:
      return x
    # dividing by the convolution of the kernel with a signal of all-ones handles correctly the lack of points at the edges (without biasing to a particular value)
    y = np.convolve(x, self.kernel, mode='same') / np.convolve(np.ones_like(x), self.kernel, mode='same')
    if len(self.kernel) > len(x):  # crop if larger (happens when filter is larger than signal, see np.convolve)
      start = len(y) // 2 - len(x) // 2
      y = y[start : start + len(x)]
    return y




"""Pure Python dashboard for monitoring deep learning experiments."""

# expose tshow utility function
from .tshow import tshow

__all__ = ['tshow']

# expose logger on import
try:
  from overboard_logger import Logger, get_timestamp, get_timestamp_folder
  __all__.extend(['Logger', 'get_timestamp', 'get_timestamp_folder'])
except ImportError:
  pass




import math, logging
from collections import OrderedDict
from importlib.util import spec_from_loader, module_from_spec

from fs import open_fs  # pyfilesystem
from fs.errors import FSError

from PyQt5.QtCore import QThread, QObject, pyqtSignal, pyqtSlot, QTimer
import pyqtgraph as pg
import pyqtgraph.opengl as gl


logger = logging.getLogger('overboard.vis')


try:
  from torch import load as pt_load
  def load(file):  # ensure tensors are not loaded on the GPU
    return pt_load(file, map_location='cpu')
except ImportError:
  # fall back to regular pickle if pytorch not installed
  from pickle import load

try:
  from torchvision.utils import make_grid
except ImportError:
  make_grid = None

try:
  import PyQt5  # needed for matplotlib to recognize the binding to this backend
  import matplotlib
  from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
  from matplotlib.pyplot import cm as colormaps
except:
  def FigureCanvas(fig):
    logger.warning("Could not load MatPlotLib.")
    return pg.PlotWidget()
  colormaps = None
  matplotlib = None

from .tshow import tshow


class Visualizations(QObject):
  """Custom visualizations, supports both MatPlotLib (MPL) and PyQtGraph (PG) figures"""

  # signal to select a different folder for the VisualizationsLoader object,
  # in a different thread
  select_folder = pyqtSignal(str, bool)

  def __init__(self, window, mpl_dpi, poll_time, log_level):
    super().__init__()

    # set logging messages threshold level
    logger.setLevel(getattr(logging, log_level.upper(), None))

    self.window = window
    window.visualizations = self  # back-reference

    self.panels = OrderedDict()  # widgets containing plots, indexed by name
    self.modules = {}  # loaded modules with visualization functions

    self.folder = None

    if matplotlib is not None:
      matplotlib.rcParams['figure.dpi'] = mpl_dpi
    
    # create loader object and thread
    if poll_time > 0:
      self.loader = VisualizationsLoader(poll_time=poll_time)
      self.thread = QThread()

      # connect VisualizationsLoader's signal to Visualizations method slot,
      # to return new visualizations
      self.loader.visualization_ready.connect(self.on_visualization_ready)
      self.loader.moveToThread(self.thread)  # move the loader object to the thread
      self.thread.start()  # start thread. note only select_folder will start the polling.

      self.select_folder.connect(self.loader.select_folder)

  def render_visualization(self, name, data, source_code):
    """Render a visualization to a MatPlotLib or PyQtGraph figure, by calling a
    custom function loaded from a pickle file"""

    # visualization function info and arguments
    func_name = data['func']
    args = data['args']
    kwargs = data['kwargs']
    source_file = data['source']

    logger.debug(f"Vis main thread: rendering {func_name} from {name}")

    # call a built-in function, e.g. simple tensor visualization
    if source_file == 'builtin' and func_name == 'tensor':
      panels = tshow(*args, **kwargs, create_window=False, title=name)
    else:
      # custom visualization function. load the saved source code file from
      # the experiment directory, and call the specified function in it.
      panels = []

      try:
        if name in self.modules:
          module = self.modules[name]  # reuse cached module
          logger.debug("Vis main thread: reused cached module")
        else:
          # create an empty module, and populate it with exec on the source code string
          module = module_from_spec(spec_from_loader(name, loader=None, origin=source_file))
          exec(source_code, module.__dict__)
          logger.debug("Vis main thread: loaded new module")

        try:
          # call the custom function, only if the module loaded successfully
          panels = getattr(module, func_name)(name, *args, **kwargs)
          
          # cache module if no error so far (otherwise reload next time, maybe it's fixed)
          self.modules[name] = module

        except Exception:
          logger.exception('Error executing visualization function ' + func_name + ' from ' + source_file)

      except Exception:
        logger.exception('Error loading visualization function from ' + source_file)

    # ensure a list is returned
    if not isinstance(panels, list): panels = [panels]
    return panels


  def on_visualization_ready(self, name, data, source_code, base_folder):
    """Called when an updated visualization (possibly new) has been
    loaded for the current experiment"""

    # ignore it if it's not the right base folder (happens when the user does not
    # wait for loading to finish and selects another experiment)
    if base_folder != self.folder: return

    # render the MatPlotLib/PyQtGraph figures
    new_plots = self.render_visualization(name, data, source_code)
    
    # assign each plot to a new or reused panel.
    # NOTE: most of this complicated logic is to deal with a specific MatPlotLib/
    # FigureCanvas bug. we cannot delete a FigureCanvas and assign a new one to the
    # same figure, or there are many graphical glitches (especially related to DPI).
    # to avoid this, we reuse the same widget (actually the parent panel widget) for
    # the same MatPlotLib figure every time (stored as overboard_panel attribute).
    new_panels = []
    old_panels = self.panels.get(name, [])
    old_panels_pg = [p for p in old_panels if p.plot_type == 'PlotItem']
    old_panels_gl = [p for p in old_panels if p.plot_type == 'GLViewWidget']

    for plot in new_plots:
      plot_type = self.get_plot_type(plot)
      if plot_type == 'Figure':  # MatPlotLib Figure
        if hasattr(plot, 'overboard_panel'):  # always reuse a previous panel
          panel = self.window.add_panel(plot.overboard_panel, name, reuse=True)
          panel.plot_widget.draw()  # ensure the figure is redrawn
        else:  # it's new
          widget = FigureCanvas(plot)
          panel = self.window.add_panel(widget, name)
          plot.overboard_panel = panel  # always associate the same panel with this figure

      elif plot_type == 'PlotItem':  # PyQtGraph PlotItem
        if old_panels_pg:
          # we can reuse an old panel and the pg.GraphicsLayoutWidget that it contains
          panel = old_panels_pg.pop()
          panel.plot_widget.clear()
          panel.plot_widget.addItem(plot)
          panel = self.window.add_panel(panel, name, reuse=True)
        else:  # it's new
          widget = pg.GraphicsLayoutWidget()
          widget.addItem(plot)
          panel = self.window.add_panel(widget, name)

      elif plot_type == 'GLViewWidget':  # PyQtGraph GLViewWidget
        widget = plot
        if old_panels_gl:
          # we can reuse an old panel, but assign the new GLViewWidget to it
          panel = old_panels_gl.pop()

          # remove the old one
          panel.plot_widget.setParent(None)
          panel.layout().removeWidget(panel.plot_widget)
          if panel.plot_widget is not widget:  # don't delete if the same widget was returned by the user's custom function
            panel.plot_widget.deleteLater()

          # insert the new
          panel.plot_widget = widget
          panel.layout().addWidget(widget, stretch=1)
          panel = self.window.add_panel(panel, name, reuse=True)
        else:  # it's new
          panel = self.window.add_panel(widget, name)

        plot.show()  # ensure the OpenGL plot is updated

      panel.plot_type = plot_type  # remember the plot type (regardless of nested widgets)
      new_panels.append(panel)
    
    # remove any panels we did not reuse from the layout
    self.delete_vis_panels(list(set(old_panels) - set(new_panels)))
    self.panels[name] = new_panels

  def select(self, exp):
    """Select a new experiment, showing its visualizations (and removing previously selected ones)"""
    # start loading visualizations
    if exp is not None:
      self.folder = exp.directory
      self.select_folder.emit(exp.directory, exp.done)
    else:
      self.folder = None
      self.select_folder.emit('', False)

    # remove previous widgets
    self.delete_vis_panels(self.all_panels())
    self.panels = {}
    self.modules = {}  # always reload modules when selecting, in case they're stale

  def all_panels(self):
    """Flatten nested list of panels"""
    if len(self.panels) == 0: return []  # special case
    return [p for panels in self.panels.values() for p in panels]
    
  def delete_vis_panels(self, panels):
    """Remove panels from the layout"""
    # MatPlotLib panels cannot be deleted, they need to be reused if the same figure is
    # displayed. note self.panels is not updated.
    for panel in panels:
      panel.setParent(None)
      if panel.plot_type != 'Figure':
        panel.deleteLater()

  def get_plot_type(self, plot):
    """Return the class name of a plot type (Figure, PlotItem or GLViewWidget).
    Throw an error if the type is not valid."""
    if type(plot).__name__ == 'Figure':
      return 'Figure'
    if isinstance(plot, pg.PlotItem):
      return 'PlotItem'
    if isinstance(plot, gl.GLViewWidget):
      return 'GLViewWidget'
    raise TypeError("Visualization functions (Logger.visualize) should return a list of "
      "MatPlotLib Figure, PyQtGraph PlotItem, or PyQtGraph GLViewWidget. A plot with "
      "class " + type(plot).__name__ + " was found.")


class VisualizationsLoader(QObject):
  """Waits for and loads new visualizations asynchronously, on a separate thread"""

  # signal to return new visualizations to the Visualizations object, on the main thread
  visualization_ready = pyqtSignal(str, dict, str, str)

  def __init__(self, poll_time):
    super().__init__()
    self.poll_time = poll_time
    self.timer = None
    self.fs = None

  @pyqtSlot(str, bool)
  def select_folder(self, folder, exp_done):
    """Monitor a different folder (or None)"""

    self.base_folder = folder
    self.exp_done = exp_done
    self.known_file_sizes = {}
    self.files_iterator = None
    self.retry_file = None
    self.source_code = {}

    if self.fs:
      self.fs.close()
      self.fs = None

    if self.timer is None:
      # create timer for the first time. note this must be done after
      # the object was moved to the thread, not in the constructor.
      self.timer = QTimer()
      self.timer.timeout.connect(self.poll)
      self.timer.setSingleShot(True)

    self.timer.stop()  # reset timer
    if self.base_folder is not None:  # start polling for changes/loading visualizations
      self.timer.start(100)

  @pyqtSlot()
  def poll(self):
    """Check for new or updated visualizations.
    Since the main use case involves remote files mounted with SSHFS/NFS, polling is
    the only viable mechanism to detect changes. This is further argued here:
    https://github.com/samuelcolvin/watchgod#why-no-inotify--kqueue--fsevent--winapi-support"""

    if not self.base_folder: return

    if not self.fs:  # create a file system object for the visualizations folder
      try:
        self.fs = open_fs(self.base_folder + '/visualizations')
      except FSError:
        # directory doesn't exist yet, try again later
        self.timer.start(self.poll_time * 1000)
        return

    # find files in the visualizations directory
    if self.files_iterator is None:
      self.files_iterator = self.fs.filterdir('.', files=['*.pth'], namespaces=['details'])

    if self.retry_file:  # try the same file again if required
      entry = self.retry_file
      self.retry_file = None
    else:
      # get next file
      try:
        entry = next(self.files_iterator)
      except StopIteration:
        entry = None

    # get pytorch pickle files
    if entry:
      name = entry.name[:-4]  # remove extension
      new_size = entry.size

      if new_size != self.known_file_sizes.get(name):
        # new file or file size changed
        self.known_file_sizes[name] = new_size

        # if the source code hasn't been loaded yet, read it
        if name not in self.source_code:
          try:
            self.source_code[name] = self.fs.readtext(name + '.py')
          except FSError:  # not found, must be a built-in (like tshow)
            self.source_code[name] = None

        # load the file (asynchronously with the main thread)
        try:
          with self.fs.open(name + '.pth', mode='rb') as file:
            data = load(file)

          if not isinstance(data, dict) or 'func' not in data:
            raise OSError("Attempted to load a visualization saved with a different protocol version (saving with PyTorch and loading without it is not supported, and vice-versa).")

          # send a signal with the results to the main thread
          self.visualization_ready.emit(name, data, self.source_code[name], self.base_folder)

        except Exception as err:
          # ignore errors about incomplete data, since file may
          # still be written to; otherwise log the error.
          if isinstance(err, RuntimeError) and 'storage has wrong size' in str(err):
            self.retry_file = entry  # try this file again later
          else:
            logger.exception(f"Error loading visualization data from {self.base_folder}/{name}.pth")
    
    # wait a bit before checking next file, or a longer time if finished all files.
    # if the experiment is done, don't check again at the end.
    if entry:
      self.timer.start(100)
    elif not self.exp_done:
      self.files_iterator = None  # check directory contents from scratch next time
      self.timer.start(self.poll_time * 1000)




#os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "TRUE"

import PyQt5.QtWidgets as QtWidgets
import PyQt5.QtCore as QtCore
import PyQt5.QtGui as QtGui
from PyQt5.QtCore import Qt

# needed right after QT imports for high-DPI screens
#QtWidgets.QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
#QtWidgets.QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)

import os, json
from time import time
from datetime import datetime, timezone
from functools import partial

import pyqtgraph as pg

from .flowlayout import FlowLayout
from .fastslider import Slider
#from .plots import Smoother


filter_tooltip_text = """Write a Python expression and then press Enter to filter.

Example: regularization >= 0.1 and batch_norm == True

Runs (displayed as table rows) for which the expression evaluates to False will be hidden.
Any hyper-parameters (displayed as table column headers) can be used in this expression.
Hyper-parameters are saved automatically by passing them to the Logger instance that records a run's results (see Logger documentation)."""


# constants referring to different data types stored in table cells.
# e.g.: item.data(table_data_exp), with item a QTableWidgetItem
table_data_exp = Qt.UserRole  # experiment name, identifies table row
table_data_column = Qt.UserRole + 1  # column name
table_data_order = Qt.UserRole + 2  # sorting order (for custom sorting, used for datetimes)


class Window(QtWidgets.QMainWindow):
  def __init__(self, max_hidden_history, window_title, clear_settings):
    super(Window, self).__init__(parent=None)
    
    self.experiments = None  # object that manages experiments
    self.plots = None  # object that manages plots
    self.visualizations = None  # object that manages custom visualizations
    self.last_process_events = time()  # to update during heavy loads
    self.rebuilding_plots = False  # used by rebuild_plots
    self.max_hidden_history = max_hidden_history
    self.selected_first_exp = False  # select first experiment to load meta, for discoverability

    # persistent settings
    self.settings = QtCore.QSettings('OverBoard', 'OverBoard')
    if clear_settings:
      self.settings.clear()  # clean slate

    # get screen size
    screen_size = QtWidgets.QDesktopWidget().availableGeometry(self).size()

    # create the parent of both sidebar and plot area
    main = QtWidgets.QSplitter(self)

    # create sidebar
    sidebar = QtWidgets.QGridLayout()
    sidebar.setAlignment(Qt.AlignTop)
    widget = QtWidgets.QWidget(main)  # need dummy widget to wrap box layout
    widget.setLayout(sidebar)

    # panel size slider
    panel_size = float(self.settings.value('panel_size', screen_size.width() * 0.2))
    sidebar.addWidget(QtWidgets.QLabel('Panel size'), 0, 0)
    slider = Slider(Qt.Horizontal)  # replaces QtWidgets.QSlider(Qt.Horizontal)
    slider.setMinimum(screen_size.width() * 0.05)
    slider.setMaximum(screen_size.width() * 0.8)
    slider.setTickInterval(screen_size.width() * 0.005)
    slider.setValue(panel_size)  # initial value
    slider.valueChanged.connect(self.on_size_slider_changed)
    sidebar.addWidget(slider, 0, 1)
    self.size_slider = slider
    
    # dropdown lists for plot configuration
    self.x_dropdown = self.create_dropdown(sidebar, label='X axis', default='iteration',
      options=['Panel metric', 'All metrics', 'iteration', 'time', 'time (relative)'], setting_name='x_dropdown')

    self.x_categorical_checkbox = self.create_checkbox(sidebar,
      label='Treat X as categorical', default=False, setting_name='x_categorical_checkbox')

    self.y_dropdown = self.create_dropdown(sidebar, label='Y axis', default='Panel metric',
      options=['Panel metric', 'All metrics', 'iteration', 'time', 'time (relative)'], setting_name='y_dropdown')

    self.y_categorical_checkbox = self.create_checkbox(sidebar,
      label='Treat Y as categorical', default=False, setting_name='y_categorical_checkbox')

    self.metrics_subset_dropdown = self.create_dropdown(sidebar,
      label='Metrics subset', options=[],
      setting_name='metrics_subset_dropdown', checkable=True)

    self.panel_dropdown = self.create_dropdown(sidebar, label='Panels', default='One per metric',
      options=['Single panel', 'One per metric', 'One per run'], setting_name='panel_dropdown')

    self.scalar_dropdown = self.create_dropdown(sidebar, label='Scalar display', default='Last value',
      options=['Last value', 'Maximum', 'Minimum'], setting_name='scalar_dropdown')

    self.merge_dropdown = self.create_dropdown(sidebar, label='Merge', default='Nothing',
      options=['Nothing'], setting_name='merge_dropdown', reset_style=True)

    (self.merge_line_dropdown, ml_label) = self.create_dropdown(sidebar,
      label='Merged line', default='Mean', options=['Mean', 'Median'],
      setting_name='merge_line_dropdown', return_label=True)

    (self.merge_shade_dropdown, ms_label) = self.create_dropdown(sidebar,
      label='Merged shade', default='2 x standard deviations',
      options=[f"{i} x standard deviation" for i in range(1, 4)] + ['Maximum and minimum'], setting_name='merge_shade_dropdown', return_label=True)

    # widgets to hide/show (merge options) depending on whether merging is on
    self.merge_option_widgets = [ml_label, self.merge_line_dropdown,
      ms_label, self.merge_shade_dropdown]
    self.merge_dropdown.activated.connect(self.on_merge_dropdown_activated)
    self.on_merge_dropdown_activated()  # merge options may start hidden

    # experiments filter text box
    rows = sidebar.rowCount()
    sidebar.addWidget(QtWidgets.QLabel('Filter'), rows, 0)
    edit = QtWidgets.QLineEdit(self.settings.value('filter_edit', ''))
    edit.setPlaceholderText('Hover for help')
    edit.returnPressed.connect(self.on_filter_ready)
    edit.setToolTipDuration(60000)  # 1 minute
    edit.setToolTip(filter_tooltip_text)
    edit.focusInEvent = self.on_filter_focus
    sidebar.addWidget(edit, rows, 1)
    self.filter_edit = edit
    self.compiled_filter = None  # compiled code of filter

    # filter auto-complete
    history = self.settings.value('filter_completer', None)
    if history is None: history = []  # QSettings doesn't like this empty list as default value
    completer = QtWidgets.QCompleter(history)
    edit.setCompleter(completer)

    """# smoothness slider
    sidebar.addWidget(QtWidgets.QLabel('Smoothness'))
    slider = Slider(Qt.Horizontal)  #QtWidgets.QSlider(Qt.Horizontal)
    slider.setMinimum(0)
    slider.setMaximum(10 * 4)  # smooth_slider_changed always divides by 4 (since sliders only support ints)
    slider.setTickInterval(1)
    slider.setValue(args.smoothen)  # initial value
    slider.valueChanged.connect(self.smooth_slider_changed)
    sidebar.addWidget(slider)
    self.smooth_slider = slider
    self.smoother = Smoother(args.smoothen)"""

    # experiments list in sidebar, as a table
    table = QtWidgets.QTableWidget(0, 4)  # start with zero rows
    table.setHorizontalHeaderLabels(['', 'run', 'timestamp', 'notes'])  # initial column headers

    self.table_col_by_name = {'timestamp': 2, 'notes': 3}  # maps column names (hyper-parameters) to column indices
    table.sortItems(2, Qt.DescendingOrder)  # sort by timestamp column

    # table style
    table.setShowGrid(False)
    table.setAlternatingRowColors(True)

    table.setSelectionBehavior(QtWidgets.QTableView.SelectRows)  # allow selecting rows
    table.setSelectionMode(QtWidgets.QTableView.SingleSelection)
    table.setAutoScroll(False)  # don't scroll when user clicks table cells

    table.setEditTriggers(QtWidgets.QAbstractItemView.AllEditTriggers)  # start editing cell with first click
    table.itemChanged.connect(self.on_table_edit)
    self.table_edit_enabled = True

    table.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
    table.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)  # smooth scrolling
    table.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
    table.setItemDelegate(ElidedTextItemDelegate())  # fix bug where text is not elided properly
    
    table.verticalHeader().hide()  # hide vertical header
    
    header = table.horizontalHeader()  # configure horizontal header
    header.setStretchLastSection(True)
    header.setDefaultAlignment(Qt.AlignLeft)
    header.setSectionsMovable(True)
    #header.setTextElideMode(Qt.ElideRight)  # show ellipsis for cut-off headers. note: takes too much space.
    header.setHighlightSections(False)  # no bold header when cells are clicked

    table.itemSelectionChanged.connect(self.on_table_select)
    table.mousePressEvent = self.on_table_click
    table.contextMenuEvent = self.on_table_context_menu

    self.table = table
    self.selected_exp = (None, None)
    sidebar.addWidget(table, sidebar.rowCount(), 0, 1, 2)
    
    # create the scroll area with plots
    (plot_scroll_widget, self.scroll_area) = create_scroller()
    main.addWidget(self.scroll_area)

    # override wheel event to prevent a bug
    self.scroll_area.wheelEvent = self.scroll_wheel_event

    # finish the parent: let sidebar width remain fixed when resizing the window, while
    # the plot area can vary; and set initial sidebar size.
    main.setStretchFactor(0, 0)
    main.setStretchFactor(1, 1)
    sidebar_size = screen_size.width() * 0.15
    main.setSizes([sidebar_size, 1])

    # main layout for plots
    self.flow_layout = FlowLayout(plot_scroll_widget)
    
    self.setCentralWidget(main)

    # window size and title
    self.resize(screen_size.width() * 0.6, screen_size.height() * 0.95)
    self.setWindowTitle(window_title)

    self.clipboard = QtWidgets.QApplication.clipboard()

    # load list of previously-hidden experiments. this is looked up when an
    # experiment loads. use a dict to preserve order and have fast lookups.
    self.hidden_exp_paths = self.settings.value('hidden_exp_paths', None)
    if self.hidden_exp_paths is None:
      self.hidden_exp_paths = []
    self.hidden_exp_paths = {e: None for e in self.hidden_exp_paths}

    # compile loaded filter
    self.on_filter_ready()

  def process_events_if_needed(self):
    """Process events if enough time has passed, to keep the GUI
    responsive during potentially heavy operations in the main thread"""
    if time() - self.last_process_events > 0.5:  # limit to once every 0.5 seconds
      QtWidgets.QApplication.processEvents()
      self.last_process_events = time()

  def add_panel(self, widget, title, add_to_layout=True, reuse=False):
    # adds a panel to the FlowLayout (main plots display), containing a widget (e.g. FigureCanvas).
    # first, create a QtWidget+QVBoxLayout around it, to show the title
    if not reuse:
      label = QtWidgets.QLabel(title)
      label.setAlignment(Qt.AlignCenter)
      label.mousePressEvent = lambda _: self.on_panel_title_click(panel)

      vbox = QtWidgets.QVBoxLayout()
      vbox.addWidget(label, stretch=0)
      vbox.addWidget(widget, stretch=1)

      panel = QtWidgets.QWidget()
      panel.setLayout(vbox)
      panel.plot_widget = widget  # keep a reference to the inner widget
      panel.title_widget = label
    else:
      panel = widget  # reusing a previous panel (less common)
      panel.title_widget.setText(title)
    
    # set the size
    panel_size = self.size_slider.value()
    panel.setFixedSize(panel_size, panel_size)

    if add_to_layout and not reuse:  # add to window's flow layout
      self.flow_layout.addWidget(panel)

    self.sort_panels()  # sort to place this panel in the right place
    return panel
  
  def on_panel_title_click(self, panel):
    """Hide or show a panel's content, when its title text is clicked"""
    if panel.plot_widget.isVisible():
      panel.plot_widget.hide()
      panel.setFixedHeight(2 * panel.title_widget.height())
      panel.title_widget.setText(panel.title_widget.text() + ' (hidden)')
      panel.title_widget.setStyleSheet('color: #909090')
    else:
      panel.plot_widget.show()
      panel.setFixedHeight(self.size_slider.value())
      panel.title_widget.setText(panel.title_widget.text()[:-len(' (hidden)')])
      panel.title_widget.setStyleSheet('')
    self.sort_panels()  # move the hidden panel to the end, or back again

  def sort_panels(self):
    """Sort the panels alphabetically by title (and some other rules)"""
    self.flow_layout._items.sort(key=get_panel_sort_key)

  def create_dropdown(self, sidebar, label, options, setting_name, default='', reset_style=False, checkable=False, return_label=False):
    """Create a new dropdown menu, associated with a persistent setting"""
    rows = sidebar.rowCount()
    label = QtWidgets.QLabel(label)
    sidebar.addWidget(label, rows, 0)

    if not checkable:
      dropdown = QtWidgets.QComboBox()
    else:
      dropdown = CheckableComboBox()

    for (idx, option) in enumerate(options):
      dropdown.addItem(option)

    if not checkable:
      default = self.settings.value(setting_name, default, type=str)
      if default:
        dropdown.setCurrentText(default)
    
    signal = partial(self.rebuild_plots, reset_style=reset_style)
    if not checkable:
      dropdown.activated.connect(signal)
    else:
      dropdown.view().clicked.connect(signal)

    sidebar.addWidget(dropdown, rows, 1)
    return (dropdown if not return_label else (dropdown, label))

  def create_checkbox(self, sidebar, label, setting_name, default):
    """Create a new checkbox, associated with a persistent setting"""
    rows = sidebar.rowCount()
    checkbox = QtWidgets.QCheckBox(label)
    checkbox.setChecked(self.settings.value(setting_name, default, type=bool))
    checkbox.toggled.connect(self.rebuild_plots)
    sidebar.addWidget(checkbox, rows, 0, 1, 2)
    return checkbox


  def on_exp_init(self, exp):
    """Called by Experiment when it is initialized"""
    # add experiment to table
    table = self.table

    # disable sorting to prevent bug when adding items, to restore afterwards
    header = table.horizontalHeader()
    prev_sort = (header.sortIndicatorSection(), header.sortIndicatorOrder())
    table.setSortingEnabled(False)

    # new row
    row = table.rowCount()
    table.insertRow(row)

    # persistent mapping between an experiment and its row (even as they're sorted)
    exp.table_row = QtCore.QPersistentModelIndex(table.model().index(row, 0))

    # create icon (empty label with writable pixmap)
    size = table.rowHeight(row)
    pixmap = QtGui.QPixmap(size, size)  # make it square
    icon = QtWidgets.QLabel()
    icon.setPixmap(pixmap)
    icon.setFixedWidth(size)
    icon.mousePressEvent = lambda _: self.on_icon_click(icon, exp)

    table.setCellWidget(row, 0, icon)
    self.redraw_icon(exp, icon)

    # show experiment name
    self.set_table_cell(row, 1, exp.name, exp, 'run')

    # restore sorting, and sort state (which column and direction)
    table.setSortingEnabled(True)
    table.sortItems(*prev_sort)

    self.process_events_if_needed()

  def redraw_icon(self, exp, icon=None, style=None):
    """Update an icon in the table by redrawing its pixmap with an experiment's style.
    Both the icon widget and the style are optional (for optimization only)."""
    if icon is None:  # allow omitting the icon widget for convenience
      icon = self.table.cellWidget(exp.table_row.row(), 0)

    pixmap = icon.pixmap()
    pixmap.fill()  # white background by default

    (x, y, w, h) = pixmap.rect().getRect()
    painter = QtGui.QPainter(pixmap)
    painter.fillRect(x, y, w, h, QtGui.QColor('white'))

    # draw box around icon
    pen = QtGui.QPen()
    pen.setWidth(2)
    if exp.is_selected:
      pen.setWidth(4)
    else:
      pen.setWidth(2)
    pen.setColor(QtGui.QColor("#EAEAF2"))
    painter.setPen(pen)

    painter.drawRoundedRect(x + 0.1 * w, y + 0.1 * h, 0.8 * w, 0.8 * h, 0.15 * w, 0.15 * h)

    # draw line, if the experiment is visible
    if exp.visible:
      if style is None:
        style = self.plots.get_exp_style(exp, assign=False)
      if style is not None:  # may not be assigned yet
        pen = pg.mkPen(style)
        if exp.is_selected:
          pen.setWidth(pen.width() + 2)
        painter.setPen(pen)
        painter.drawLine(x + 0.2 * w, y + 0.5 * h, x + 0.8 * w, y + 0.5 * h)

    painter.end()
    icon.repaint()  # important, to show changes on screen

  def on_exp_meta_ready(self, exp):
    """Called by Experiment when the meta-data has been read"""

    # print a row of meta-data (argument) values for this experiment in the table
    (table, table_col_by_name) = (self.table, self.table_col_by_name)
    row = exp.table_row.row()
    added_columns = False
    self.table_edit_enabled = False  # avoid on_table_edit when filling in an editable cell for the first time

    # disable sorting to prevent bug when adding items, to restore afterwards
    header = table.horizontalHeader()
    prev_sort = (header.sortIndicatorSection(), header.sortIndicatorOrder())
    table.setSortingEnabled(False)

    for arg_name in exp.meta.keys():
      if not arg_name.startswith('_'):
        if arg_name not in table_col_by_name:  # a new argument name, add a column
          col = table.columnCount()
          table.setColumnCount(col + 1)
          table.setHorizontalHeaderItem(col, QtWidgets.QTableWidgetItem(arg_name))  # , QtWidgets.QTableWidgetItem.Type
          table_col_by_name[arg_name] = col
          added_columns = True
        else:
          col = table_col_by_name[arg_name]
        
        cell_value = exp.meta.get(arg_name, '')
        
        self.set_table_cell(row, col, cell_value, exp, arg_name, editable=(arg_name == 'notes'))

    if 'notes' not in exp.meta:  # explicitly create editable notes cell, if not created already
      self.set_table_cell(row, table_col_by_name['notes'], '', exp, 'notes', editable=True)

    if added_columns:
      self.resize_table()

    # update dropdown lists to include all hyper-parameter names
    for arg_name in exp.meta.keys():
      for widget in [self.x_dropdown, self.y_dropdown, self.panel_dropdown, self.merge_dropdown]:
        if widget.findText(arg_name) < 0:
          widget.addItem(arg_name)
    
    # hide row if filter says so
    self.filter_experiment(exp)

    self.table_edit_enabled = True

    # restore sorting, and sort state (which column and direction)
    table.setSortingEnabled(True)
    table.sortItems(*prev_sort)

    # select first visible experiment, for discoverability
    if exp.is_visible() and not self.selected_first_exp:
      self.selected_first_exp = True
      self.select_experiment(exp)

    self.process_events_if_needed()

  def on_exp_header_ready(self, exp):
    """Called by Experiment when the header data (metrics/column names) has been read"""
    # update dropdown lists to include all metric names
    unchecked = self.settings.value('metrics_subset_dropdown', [])
    for name in exp.metrics:
      if self.x_dropdown.findText(name) < 0:
        self.x_dropdown.addItem(name)
      if self.y_dropdown.findText(name) < 0:
        self.y_dropdown.addItem(name)
      if self.metrics_subset_dropdown.findText(name) < 0:
        self.metrics_subset_dropdown.addItem(name, checked=name not in unchecked)

  def rebuild_plots(self, reset_style=False):
    """Rebuild all plots (e.g. when plot options such as x/y axis change)"""
    if not self.rebuilding_plots:  # cancel when already rebuilding the plots (i.e. called recursively)
      self.rebuilding_plots = True
      if reset_style:
        self.plots.drop_all_exp_styles()
      self.plots.remove_all()
      for exp in self.experiments.exps.values():
        visible = self.plots.add(exp)
        if visible:
          self.process_events_if_needed()  # keep it responsive
      self.rebuilding_plots = False

  def set_table_cell(self, row, col, value, exp, col_name, selectable=True, editable=False):
    """Set a single table cell in the sidebar"""
    # try to interpret as integer or float, to allow numeric sorting of columns. editable is only for strings.
    if not editable and not isinstance(value, (int, float, datetime)):
      value = str(value)  # handle e.g. dicts
      try:
        value = float(value)
        if int(value) == value:  # store as int if possible, prints better
          value = int(value)
      except ValueError: pass

    if isinstance(value, str):  # faster option for strings
      item = QtWidgets.QTableWidgetItem(value)

    elif isinstance(value, datetime):  # show nicer-looking timestamps, sorted correctly
      item = SortableTableItem(print_datetime(value), value.timestamp)

    else:  # sort as number if not a string
      item = QtWidgets.QTableWidgetItem()
      item.setData(Qt.EditRole, QtCore.QVariant(value))

    item.setData(table_data_exp, exp.name)  # pointer to original experiment (by name), i.e. row name
    item.setData(table_data_column, col_name)  # column name

    flags = Qt.ItemIsEnabled
    if selectable: flags |= Qt.ItemIsSelectable
    if editable: flags |= Qt.ItemIsEditable
    item.setFlags(flags)
    
    self.table.setItem(row, col, item)
    return item

  def resize_table(self):
    """Resize the column headers to fully contain the header text
    (note that Qt's resizeColumnsToContents ignores the headers)"""
    table = self.table
    table.resizeColumnsToContents()
    
    max_width = int(0.9 * table.parentWidget().width())  # don't let any column become wider than 90% of the sidebar
    header = table.horizontalHeader()
    metrics = QtGui.QFontMetrics(table.horizontalHeaderItem(0).font())
    for col in range(table.columnCount()):
      text = table.horizontalHeaderItem(col).text()
      text_width = metrics.boundingRect(text).width()
      header.resizeSection(col, min(max_width, max(table.sizeHintForColumn(col), text_width + 20)))  # needs some extra width

  def on_icon_click(self, icon, exp):
    """Toggle visibility of a given experiment, when the
    icon (first column of table) is clicked"""

    if exp.is_filtered: return  # shouldn't happen
    if exp.visible:
      # remove all associated plots, and reset the experiment style so it can be used by others
      exp.visible = False
      self.plots.remove(exp)
      self.plots.drop_exp_style(exp)
    else:
      # create plots (a new style will be assigned if necessary)
      exp.visible = True
      self.plots.add(exp)

    # update icon
    self.redraw_icon(exp, icon)

  def on_table_select(self):
    """Select experiment on table row click"""
    exp = None
    selected = self.table.selectedItems()
    if len(selected) > 0:  # select new one
      exp = self.experiments.exps[selected[0].data(table_data_exp)]
    self.select_experiment(exp, clicked_table=True)

  def on_table_click(self, event):
    """Clear selection on click (before selecting a row), to allow de-selecting by clicking outside table items"""
    self.table.clearSelection()
    QtWidgets.QTableWidget.mousePressEvent(self.table, event)
  
  def on_table_context_menu(self, event):
    """Show menu when right-clicking a cell in the table"""
    # get the selected row and column
    row = self.table.rowAt(event.pos().y())
    col = self.table.columnAt(event.pos().x())

    action = QtWidgets.QAction('Copy', self)
    action.triggered.connect(lambda: self.copy_cell(row, col))
    
    self.menu = QtWidgets.QMenu(self)
    self.menu.addAction(action)
    self.menu.popup(QtGui.QCursor.pos())

  def copy_cell(self, row, col):
    """Copy a table cell to the clibpard (called via the context menu)"""
    cell = self.table.item(row, col)
    if cell:
      self.clipboard.setText(cell.text())

  def select_experiment(self, exp=None, clicked_table=False):
    """Select an experiment in the table, highlighting it in the plots.
    Passing None deselects previous one."""

    if isinstance(exp, str):  # experiment name, look it up
      exp = self.experiments.exps.get(exp, None)

    # unselect previous experiment first
    (old_exp, old_icon) = self.selected_exp
    if old_exp:
      old_exp.is_selected = False
      self.redraw_icon(old_exp, old_icon)
      self.plots.add(old_exp)  # update its view, in case it's visible

    if exp and exp.is_visible():  # don't select if invisible
      icon = self.table.cellWidget(exp.table_row.row(), 0)

      exp.is_selected = True
      self.plots.add(exp)
      self.selected_exp = (exp, icon)
      self.visualizations.select(exp)
      self.redraw_icon(exp, icon)

      if not clicked_table:  # update table selection
        self.table.selectRow(exp.table_row.row())
    else:
      self.selected_exp = (None, None)
      self.visualizations.select(None)
      if not clicked_table:  # update table selection
        self.table.clearSelection()

  def on_table_edit(self, item):
    """Finished editing a cell in the table"""
    if self.table_edit_enabled:
      if item.data(table_data_column) == 'notes':
        # update experiment meta-data with updated note, and update the JSON file
        exp = self.experiments.exps[item.data(table_data_exp)]

        if item.text() != exp.meta.get('notes', ''):  # user changed the text
          # write to another file and only then replace the original (in case of crashes/bad writes)
          exp.meta['notes'] = item.text()
          with open(exp.directory + '/meta.json.partial', 'w') as file:
            json.dump(exp.meta, file, sort_keys=True, indent=4, default=str)
          os.replace(exp.directory + '/meta.json.partial', exp.directory + '/meta.json')
          

  def on_filter_ready(self):
    """User pressed Enter in filter text box, filter the experiments"""
    if len(self.filter_edit.text().strip()) == 0:  # no filter
      self.compiled_filter = None
      self.filter_edit.setStyleSheet("color: black;")
    else:
      # compile filter code
      try:
        self.compiled_filter = compile(self.filter_edit.text(), '<filter>', 'eval')
        self.filter_edit.setStyleSheet("color: black;")
      except Exception as err:
        self.show_filter_error(err)
        return

    if self.experiments is not None:
      for exp in self.experiments.exps.values():
        err = self.filter_experiment(exp)
        if err: return
        self.process_events_if_needed()  # keep it responsive
      
      # add to auto-complete model, if there was no error
      text = self.filter_edit.text()
      model = self.filter_edit.completer().model()
      entries = model.stringList()
      if text and text not in entries:
        entries.insert(0, text)  # insert at top of list
        model.setStringList(entries)

  def filter_experiment(self, exp):
    """Apply filter to a single experiment, hiding it or showing it"""
    if self.compiled_filter is None:
      # no filter, show experiment unconditionally
      if exp.is_filtered:
        self.table.setRowHidden(exp.table_row.row(), False)
        exp.is_filtered = False
        self.plots.add(exp)
    else:
      # create a dict with the hyper-parameters from this experiment, and all
      # missing hyper-parameters set to None, to be accessed by the filter function.
      vars_table = dict(zip(self.table_col_by_name.keys(), [None] * len(self.table_col_by_name)))
      vars_table.update(exp.meta)

      # other special variables that will be available to the filter function
      vars_table['run'] = exp.name
      vars_table['datetime'] = datetime  # useful to manipulate timestamps
      vars_table['timezone'] = timezone

      # evaluate filter to obtain boolean
      try:
        hide = not eval(self.compiled_filter, vars_table, vars_table)
      except Exception as err:
        self.show_filter_error(err)
        return True

      # hide or show depending on context
      was_hidden = not exp.is_visible()
      exp.is_filtered = hide  # will be checked by Plots.add

      self.table.setRowHidden(exp.table_row.row(), hide)
      if hide:
        if not was_hidden:
          self.plots.remove(exp)
      else:
        self.plots.add(exp)
      return False

  def show_filter_error(self, err):
    """Show filter expression error as a tooltip, and change color to red"""
    text = err.__class__.__name__ + ": " + str(err)
    QtWidgets.QToolTip.showText(QtGui.QCursor.pos(), text, self.filter_edit)
    self.filter_edit.setStyleSheet("color: #B00000;")

  def on_filter_focus(self, event):
    """Event handler for when filter line-edit widget gets focus"""
    QtWidgets.QLineEdit.focusInEvent(self.filter_edit, event)
    self.filter_edit.completer().complete()

  def on_size_slider_changed(self):
    """Resize panels for plots and visualizations"""
    panel_size = self.size_slider.value()
    for panel in self.plots.panels.values():
      panel.setFixedSize(panel_size, panel_size)
    for panel_group in self.visualizations.panels.values():
      for panel in panel_group:
        panel.setFixedSize(panel_size, panel_size)
  
  #def smooth_slider_changed(self):
  #  self.smoother = Smoother(self.smooth_slider.value() / 4.0)

  def on_merge_dropdown_activated(self):
    """Show/hide other merge options depending on whether merging is on"""
    visibility = (self.merge_dropdown.currentText() != 'Nothing')
    for widget in self.merge_option_widgets:
      widget.setVisible(visibility)

  def scroll_wheel_event(self, event):
    """Override QScrollArea wheelEvent, to prevent it from scrolling simultaneously with PyQtGraph's PlotItem"""
    # if one of the panels/plots of the flow layout is under
    # the mouse, don't let QScrolArea scroll too
    for widget_item in self.flow_layout._items:
      if widget_item.widget().underMouse():
        event.ignore()
        return
    QtWidgets.QScrollArea.wheelEvent(self.scroll_area, event)
  
  def closeEvent(self, event):
    """Write state to settings before closing"""
    self.settings.setValue('panel_size', self.size_slider.value())
    self.settings.setValue('x_dropdown', self.x_dropdown.currentText())
    self.settings.setValue('x_categorical_checkbox', self.x_categorical_checkbox.isChecked())
    self.settings.setValue('y_dropdown', self.y_dropdown.currentText())
    self.settings.setValue('y_categorical_checkbox', self.y_categorical_checkbox.isChecked())
    self.settings.setValue('panel_dropdown', self.panel_dropdown.currentText())
    self.settings.setValue('scalar_dropdown', self.scalar_dropdown.currentText())
    self.settings.setValue('merge_dropdown', self.merge_dropdown.currentText())
    self.settings.setValue('merge_line_dropdown', self.merge_line_dropdown.currentText())
    self.settings.setValue('merge_shade_dropdown', self.merge_shade_dropdown.currentText())

    self.settings.setValue('filter_edit', self.filter_edit.text())

    # save unchecked metric subset, up to 50 entries
    checked = set(self.metrics_subset_dropdown.get_checked_list())
    prev_unchecked = self.settings.value('metrics_subset_dropdown', [], type=list)  # previously hidden
    unchecked = {m: None for m in prev_unchecked if m not in checked}  # previously hidden but now checked again
    unchecked.update({m: None for m in self.metrics_subset_dropdown.get_checked_list(unchecked=True)})  # merge with newly hidden
    if len(unchecked) > 50:  # keep N last
      unchecked = list(unchecked)[-50:]
    self.settings.setValue('metrics_subset_dropdown', list(unchecked))

    # save auto-complete model for filter, up to 50 entries
    m = self.filter_edit.completer().model()
    history = [m.data(m.index(i), 0) for i in range(min(50, m.rowCount()))]
    self.settings.setValue('filter_completer', history)

    # save hidden status of individual experiments, preserving order
    self.hidden_exp_paths.update({exp.directory: None for exp in self.experiments.exps.values() if not exp.visible})
    if len(self.hidden_exp_paths) > self.max_hidden_history:  # keeping N last
      self.hidden_exp_paths = list(self.hidden_exp_paths)[-self.max_hidden_history:]
    self.settings.setValue('hidden_exp_paths', list(self.hidden_exp_paths))

    self.settings.sync()
    event.accept()


class SortableTableItem(QtWidgets.QTableWidgetItem):
  """Used to define custom sort order (stored as the 3rd UserRole data)
  for table cells representing objects such as datetimes"""
  def __init__(self, text, order):
    super().__init__(text)
    self.setData(table_data_order, order)

  def __lt__(self, other):
    return (self.data(table_data_order)() < other.data(table_data_order)())


class CheckableComboBox(QtWidgets.QComboBox):
  """Drop-down list with checkable items"""
  def __init__(self, parent=None):
    super().__init__(parent)
    self.setEditable(True)
    self.view().clicked.connect(self.refresh_text)

  def addItem(self, text, checked=False):
    super().addItem(text)
    item = self.model().item(self.count() - 1)
    item.setFlags(Qt.ItemIsUserCheckable | Qt.ItemIsEnabled)
    item.setCheckState(Qt.Checked if checked else Qt.Unchecked)
    self.refresh_text()

  def refresh_text(self, item=None):
    """Shows the checked items as a comma-separated list in the QComboBox"""
    # setCurrentText doesn't work for non-editable QComboBox, so do this
    line = self.lineEdit()
    line.setText(', '.join(self.get_checked_list()))
    line.setReadOnly(True)
    line.setCursorPosition(0)
  
  def get_checked_list(self, unchecked=False):
    """Return a list of strings containing the checked items"""
    model = self.model()
    items = [model.item(i) for i in range(model.rowCount())]
    state = (Qt.Unchecked if unchecked else Qt.Checked)
    return [str(item.text()) for item in items if item.checkState() == state]


class ElidedTextItemDelegate(QtWidgets.QStyledItemDelegate):
  """Helper delegate to draw proper elided text (i.e. cut off with ...) in tables"""
  def paint(self, painter, option, index):
    if not index.isValid():
      return

    # modify a copy of the options
    opt = QtWidgets.QStyleOptionViewItem(option)
    self.initStyleOption(opt, index)
    style = QtWidgets.QApplication.style()

    # reduce the rect a bit
    painter.save()
    painter.setClipRect(opt.rect)
    padding = 3
    opt.rect = opt.rect.adjusted(padding, padding, -padding, -padding)

    # elide the text as a string
    text = opt.fontMetrics.elidedText(opt.text, Qt.ElideRight, opt.rect.width())

    # draw it, with the proper role (i.e. color if selected)
    text_role = QtGui.QPalette.NoRole
    if option.state & style.State_Selected:
      text_role = QtGui.QPalette.HighlightedText
    style.drawItemText(painter, opt.rect, opt.displayAlignment,
      opt.palette, True, text, text_role)

    # draw the item again with empty text to complete it
    painter.restore()
    opt.rect = option.rect
    opt.textElideMode = Qt.ElideNone
    opt.text = ""
    style.drawControl(QtWidgets.QStyle.CE_ItemViewItem, opt, painter, opt.widget)


def create_scroller():
  scroll_area = QtWidgets.QScrollArea()
  scroll_area.setWidgetResizable(True)

  scroll_widget = QtWidgets.QWidget(scroll_area)  # additional widget to enable scrollbars
  scroll_widget.setMinimumWidth(50)
  scroll_area.setWidget(scroll_widget)
  return (scroll_widget, scroll_area)

def print_datetime(dt):
  """Prints a friendly string with a given datetime object"""
  now = datetime.now(timezone.utc)
  dt = dt.astimezone()  # convert from UTC to local time for display
  if dt.day == now.day: return f"{dt:%X}"  # time
  if (now - dt).days <= 7: return f"{dt:%a, %X}"  # weekday, time
  if dt.year == now.year: return f"{dt.day} {dt:%b, %X}"  # day month, time
  return f"{dt.day} {dt:%b %Y, %X}"  # day month year, time


def get_panel_sort_key(item):
  """Get a key from a panel for sorting.
  They are sorted by title, but we want visualizations to come second,
  and hidden/minimized panels last."""
  widget = item.widget()
  return (not widget.plot_widget.isVisible(),  # visible panels come first
    not hasattr(widget.plot_widget, 'is_experiment_plot'),  # then experiment plots
    widget.title_widget.text())  # all else being equal, sort titles alphabetically


def set_style(app):
  app.setStyle("Fusion")

  directory = os.path.dirname(os.path.realpath(__file__))

  QtGui.QFontDatabase.addApplicationFont(directory + './assets/OpenSans-Regular.ttf')
  
  with open(directory + '/style.qss', 'r') as file:
    app.setStyleSheet(file.read())




import math, logging
import pyqtgraph as pg

try:
  from torchvision.utils import make_grid
except ImportError:
  make_grid = None

import PyQt5  # needed for matplotlib to recognize the binding to this backend
try:
  from matplotlib.pyplot import cm as colormaps
except:
  colormaps = None


tshow_images = []

def tshow(tensor, create_window=True, title='Tensor', data_range=None, grayscale=False, legend=None, **kwargs):
  """Shows a PyTorch tensor (including one or more RGB images) using PyQtGraph."""

  if make_grid is None:
    raise ImportError('Could not import torchvision (from PyTorch), which is necessary for tshow.')

  tensor = tensor.detach().cpu()
  original_shape = tensor.shape

  if data_range is None:
    data_range = (tensor.min(), tensor.max())

  if len(tensor.shape) > 4:
    logging.exception('Cannot show tensors with more than 4 dimensions.')
    return None

  # insert singleton dimensions on the left to always get 4 dimensions
  while len(tensor.shape) < 4:
    tensor = tensor.unsqueeze(0)

  sh = tensor.shape
  if sh[0] == 1:  # case of 3D tensors, leave singleton dimension for color
    tensor = tensor.reshape(sh[1], 1, sh[2], sh[3])

  sh = tensor.shape
  if sh[1] in [1, 3]:
    # a linear collection of images: channels are RGB or single-channel.
    # choose number of columns such that they are laid out in a square.
    if 'nrow' not in kwargs:
      #kwargs['nrow'] = math.ceil(math.sqrt(sh[0]))
      kwargs['nrow'] = math.ceil(math.sqrt(sh[0]) * sh[2] / sh[3])
  else:
    # when channels are not RGB or single-channel, display them as columns
    tensor = tensor.reshape(sh[0] * sh[1], 1, sh[2], sh[3])
    kwargs['nrow'] = sh[1]


  # arrange into a grid
  image = make_grid(tensor, **kwargs, normalize=False, padding=0)
  image = image.permute(2, 1, 0).numpy()  # pytorch convention to numpy image convention

  # convert grayscale RGB images to colormapped images (single-channel)
  if tensor.shape[1] == 1:
    image = image[:,:,0]

  # show it
  im_item = pg.ImageItem(image)
  title = title + ' ' + str(tuple(original_shape))

  if create_window:  # stand-alone window
    win = pg.GraphicsWindow(title=title)
    plot = win.addPlot()
  else:  # plot item to return
    plot = pg.PlotItem()
  plot.addItem(im_item)

  plot.getViewBox().invertY(True)
  plot.setAspectLocked(True)
  plot.hideAxis('left')
  plot.hideAxis('bottom')

  # draw a grid
  (cell_w, cell_h) = (sh[3], sh[2])
  (w, h) = image.shape[0:2]
  for x in range(0, w + 1, cell_w):
    plot.plot([x, x], [0, h])
  for y in range(0, h + 1, cell_h):
    plot.plot([0, w], [y, y])

  if len(image.shape) == 3:
    # RGB image
    im_item.setLevels([data_range] * 3)
  else:
    # grayscale image or heatmap, set up colormap and possibly a legend
    if not grayscale and colormaps is not None:
      # use better colormap if matplotlib is available
      colormap = colormaps.viridis
      colormap._init()
      lut = (colormap._lut * 255)[:-1,:]  # remove last row
      im_item.setLookupTable(lut)
      (low_color, high_color) = (lut[0,:3], lut[-1,:3])
    else:
      lut = []
      (low_color, high_color) = ('k', 'w')

    im_item.setLevels(data_range)

    if legend or (legend is None and not grayscale):
      # create legend with max and min values
      leg = plot.addLegend(offset=(1, 1))
      leg.addItem(FilledIcon(low_color), "Min: {:.3g}".format(data_range[0]))
      leg.addItem(FilledIcon(high_color), "Max: {:.3g}".format(data_range[1]))

      # monkey-patch paint method to draw a more opaque background
      def paint(self, p, *args):
        color = pg.mkColor(pg.getConfigOption('background'))
        color.setAlpha(200)
        p.fillRect(self.boundingRect(), pg.mkBrush(color))
      leg.paint = paint.__get__(leg)

  if create_window:
    tshow_images.append(win)  # keep reference, otherwise the window will be garbage-collected
    win.show()
  else:
    return plot


class FilledIcon(pg.graphicsItems.LegendItem.ItemSample):
  """Custom legend icon, completely filled with a single color."""
  def __init__(self, color):
    super().__init__(None)
    self.color = color
    self.setFixedWidth(20)

  def paint(self, p, *args):
    p.fillRect(self.boundingRect(), pg.mkBrush(*self.color))





import setuptools

with open("README.md", "r", encoding="utf-8") as f:
  long_description = f.read()

setuptools.setup(
  name="overboard",
  version="1.0.1",
  author="Joao Henriques",
  description="Pure Python dashboard for monitoring deep learning experiments",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/jotaf98/overboard",
  packages=['overboard'],
  package_data={'overboard': ['style.qss']},
  include_package_data=True,
  python_requires='>=3.6',
  install_requires=[
    'pyqt5>=5.12',
    'pyqtgraph>=0.11,<0.12',
    'pyopengl>=3.1',
    'fs>=2.4',
    'overboard_logger>=0.7.1'
  ],
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Operating System :: OS Independent",
  ],
)
